#!/usr/bin/env python3
# Copyright (C) 2020 ASTRON (Netherlands Institute for Radio Astronomy)
# SPDX-License-Identifier: GPL-3.0-or-later

from astropy.io import fits

"""
Script to generate a couple of unit-strength
point sources in a provided template image.
"""

with fits.open("template-image.fits") as img:
    N = img[0].data.shape[-1]

    offset = N // 4

    img[0].data[:] = 0.0
    img[0].data[0, 0, offset, offset] = 1.0
    img[0].data[0, 0, offset, N - offset] = 1.0
    img[0].data[0, 0, N - offset, offset] = 1.0
    img[0].data[0, 0, N - offset, N - offset] = 1.0
    img.writeto("wsclean-I-model.fits", overwrite=True)

    img[0].data[:] = 0.0
    img[0].data[0, 0, offset, N - offset] = 1.0
    img.writeto("wsclean-Q-model.fits", overwrite=True)

    img[0].data[:] = 0.0
    img[0].data[0, 0, N - offset, offset] = 1.0
    img.writeto("wsclean-U-model.fits", overwrite=True)

    img[0].data[:] = 0.0
    img[0].data[0, 0, N - offset, N - offset] = 1.0
    img.writeto("wsclean-V-model.fits", overwrite=True)



#!/usr/bin/env python3
# Copyright (C) 2020 ASTRON (Netherlands Institute for Radio Astronomy)
# SPDX-License-Identifier: GPL-3.0-or-later

# Work-around issue https://github.com/astropy/astropy/issues/13007
from astropy.utils import iers

iers.conf.auto_download = False

import os
import numpy as np
from oskar_utils import check_tolerance, read_oskar_beams
import run_oskar_simulation
import subprocess
import everybeam as eb

"""
Compare the station response obtained with OSKAR to the station
response obtained via EveryBeam.
"""

tolerance = (
    float(os.environ["TOLERANCE"]) if "TOLERANCE" in os.environ else 0.0
)
npixels = int(os.environ["NPIXELS"]) if "NPIXELS" in os.environ else 256
data_dir = os.environ["DATA_DIR"]
freq = 50e6
station_index = 0

# Obtain result with OSKAR
run_oskar_simulation.main(npixels)
subprocess.check_call(
    [
        "add_beaminfo.py",
        "skalowmini-coef.MS",
        os.path.join(data_dir, "skalowmini-coef.tm"),
    ]
)
subprocess.check_call(
    [
        "oskar_csv_to_hdf5.py",
        os.path.join(data_dir, "skalowmini-coef.tm"),
        "oskar.h5",
    ]
)
# subprocess.check_call(["make_station_response_image", str(npixels)])

root_name = "basefunctions_050"
A = read_oskar_beams(root_name)


# Obtain results with EveryBeam

# Time and itrf vector of reference direction
# TODO: infer this from EveryBeam?
time = 4453522026
station0 = np.array([-0.4028537349, 0.7946940582, -0.4540597121])

# Antenna coordinate_system,
# TODO: infer these via EveryBeam?
# C++ counterpart is: station->GetAntenna()->coordinate_system_.axes.[p,q,r]
p = np.array([-0.8928654139, -0.4503236089, 0])
q = np.array([-0.2032142175, 0.4029167976, 0.89239119])
r = np.array([-0.4018648212, 0.7967852292, -0.451262633])

telescope = eb.load_telescope(
    "skalowmini-coef.MS", element_response_model="skala40_wave"
)
x_v = np.linspace(-1.0, 1.0, npixels)
y_v = np.linspace(-1.0, 1.0, npixels)
B = np.empty((y_v.size, x_v.size, 2, 2), dtype=np.cdouble)
B.fill(complex(np.nan, np.nan))

for i, x in enumerate(x_v):
    for j, y in enumerate(y_v):
        if (x**2 + y**2) <= 1.0:
            z = np.sqrt(1.0 - x**2 - y**2)
            direction = x * p + y * q + z * r
            B[i, j, :, :] = telescope.station_response(
                time, station_index, freq, direction, station0, rotate=False
            )

# Compare OSKAR (A) with EveryBeam (B)
check_tolerance(tolerance, A, B)



#!/usr/bin/env python3
# Copyright (C) 2020 ASTRON (Netherlands Institute for Radio Astronomy)
# SPDX-License-Identifier: GPL-3.0-or-later

import argparse
import os
import scipy.io as sio
import os.path
import numpy as np
import h5py
from tqdm import tqdm
from glob import glob
import csv

# Parse command-line arguments
parser = argparse.ArgumentParser()
parser.add_argument(
    "input_dir", type=str, help="directory containing coefficients"
)
parser.add_argument(
    "output_file", type=str, help="name of the HDF5 output file"
)
parser.add_argument("-debug", dest="debug", action="store_true")
parser.set_defaults(input_dir=".")
parser.set_defaults(output_file="oskar.h5")
parser.set_defaults(debug=False)
args = parser.parse_args()

# Read the input directory
# This directory should contain subdirectories,
# with names corresponding to frequency.
input_dir = args.input_dir
print("Reading coefficients from: ", input_dir)

# Only read element_id 0, because EveryBeam's implementation of the OSKAR model
# currently supports only a single model for all elements
element_id = 0
nr_elements = 1

files = glob(
    os.path.join(
        input_dir, "element_pattern_spherical_wave_?_t?_??_{}_*.txt"
    ).format(element_id)
)

files = [os.path.basename(f) for f in files]

freqs = sorted(set([int(f[:-4].split("_")[-1]) for f in files]))


print("Parsing frequencies %d-%d MHz" % (int(freqs[0]), int(freqs[-1])))

# Create HDF5 file
output_file = args.output_file
print("Creating HDF5 file: ", output_file)
hf = h5py.File(output_file, "w")

# Debugging
debug = args.debug
if debug:
    freqs = [freqs[0]]
    tqdm = lambda x: x

# Parse all freqs
for freq in tqdm(freqs):
    A = None

    for i, pol in enumerate(("x", "y")):
        for j, em in enumerate(("e", "m")):
            for reim, s in zip(("re", "im"), (1.0, 1.0j)):
                filename = "element_pattern_spherical_wave_{}_t{}_{}_{}_{}.txt".format(
                    pol, em, reim, element_id, freq
                )
                alpha = np.loadtxt(
                    os.path.join(input_dir, filename), delimiter=",", ndmin=2
                )
                l_max = alpha.shape[0]
                assert alpha.shape[1] == 2 * l_max + 1
                if A is None:
                    A = np.zeros(
                        (l_max, 2 * l_max + 1, 2, 2), dtype=np.complex128
                    )
                else:
                    assert A.shape == (l_max, 2 * l_max + 1, 2, 2)
                A[:, :, i, j] += alpha * s

    # Create the output matrix
    data = np.zeros((nr_elements, l_max * (l_max + 2), 4), dtype=np.complex128)

    # Fill the output matrix
    # The inner dimension are set as follows:
    # (x_te_re, x_te_im), (x_tm_re, x_tm_im),
    # (y_te_re, y_te_im), (y_tm_re, y_tm_im)

    i = 0
    for l in range(l_max):
        n = (l + 1) * 2 + 1
        data[0, i : i + n, :] = A[l, :n, :].reshape(n, 4)
        i += n

    # Add group for current frequency
    hf.create_dataset(str(freq), data.shape, data=data)

# Close HDF5 file
hf.close()

# Done
print("Done!")



#!/usr/bin/env python3
# Copyright (C) 2020 ASTRON (Netherlands Institute for Radio Astronomy)
# SPDX-License-Identifier: GPL-3.0-or-later

# Work-around issue https://github.com/astropy/astropy/issues/13007
from astropy.utils import iers

iers.conf.auto_download = False

import os
import numpy as np
import matplotlib.pyplot as plt
import run_oskar
from oskar_utils import check_tolerance, generate_oskar_csv, read_oskar_beams
import subprocess

# Check and set parameters
l_max = int(os.environ["MAX_ORDER"]) if "MAX_ORDER" in os.environ else 2
apply_transpose = (
    os.environ["APPLY_TRANSPOSE"].upper() in ("1", "ON", "TRUE")
    if "APPLY_TRANSPOSE" in os.environ
    else False
)
tolerance = (
    float(os.environ["TOLERANCE"]) if "TOLERANCE" in os.environ else 0.0
)
npixels = int(os.environ["NPIXELS"]) if "NPIXELS" in os.environ else 256

plt.figure(figsize=(10, 6))

# Freqs (in MHz)
freqs = [50]
output_dir = "telescope.tm"
element_id = 0

root_name = f"basefunctions_{freqs[0]:03}"
for em_idx in range(2):
    for basefunction_idx in range(l_max * (l_max + 2)):
        plt.clf()

        generate_oskar_csv(
            basefunction_idx, freqs, output_dir, element_id, em_idx
        )
        run_oskar.main(npixels)
        A = read_oskar_beams(root_name)

        plt.subplot(2, 4, 1)
        plt.imshow(np.abs(A[:, :, 0, 0]).T, clim=(0, 0.25), origin="lower")
        plt.colorbar()
        plt.title("abs(Etheta)")
        plt.ylabel("OSKAR")

        plt.subplot(2, 4, 2)
        plt.imshow(np.abs(A[:, :, 0, 1]).T, clim=(0, 0.25), origin="lower")
        plt.colorbar()
        plt.title("abs(Ephi)")

        plt.subplot(2, 4, 3)
        plt.imshow(
            np.angle(A[:, :, 0, 0]).T,
            clim=(-np.pi, np.pi),
            cmap="twilight",
            origin="lower",
        )
        plt.colorbar()
        plt.title("angle(Etheta)")

        plt.subplot(2, 4, 4)
        plt.imshow(
            np.angle(A[:, :, 0, 1]).T,
            clim=(-np.pi, np.pi),
            cmap="twilight",
            origin="lower",
        )
        plt.colorbar()
        plt.title("angle(Ephi)")

        l = int(np.sqrt(basefunction_idx + 1))
        m = basefunction_idx - l * l + 1 - l
        s = em_idx

        if not apply_transpose:
            generate_oskar_csv(
                basefunction_idx, freqs, output_dir, element_id, em_idx
            )
        else:
            # flip the sign of m
            generate_oskar_csv(
                l * l - 1 + l - m, freqs, output_dir, element_id, em_idx
            )

        subprocess.check_call(
            ["oskar_csv_to_hdf5.py", "telescope.tm", "oskar-comparison.h5"]
        )
        subprocess.check_call(["make_element_response_image", str(npixels)])

        B = np.load("response.npy")

        if apply_transpose:
            B *= (-1) ** (m + 1)

        plt.subplot(2, 4, 5)
        plt.imshow(np.abs(B[:, :, 0, 0]).T, clim=(0, 0.25), origin="lower")
        plt.colorbar()
        plt.title("abs(Etheta)")
        plt.ylabel("EveryBeam")

        plt.subplot(2, 4, 6)
        plt.imshow(np.abs(B[:, :, 0, 1]).T, clim=(0, 0.25), origin="lower")
        plt.colorbar()
        plt.title("abs(Ephi)")

        plt.subplot(2, 4, 7)
        plt.imshow(
            np.angle(B[:, :, 0, 0]).T,
            clim=(-np.pi, np.pi),
            cmap="twilight",
            origin="lower",
        )
        plt.colorbar()
        plt.title("angle(Etheta)")

        plt.subplot(2, 4, 8)
        plt.imshow(
            np.angle(B[:, :, 0, 1]).T,
            clim=(-np.pi, np.pi),
            cmap="twilight",
            origin="lower",
        )
        plt.colorbar()
        plt.title("angle(Ephi)")
        plt.gcf().suptitle("l = {}, m = {}, s = {}".format(l, m, s))
        plt.savefig("basefunction{}-{}".format(basefunction_idx, em_idx))

        check_tolerance(tolerance, A, B)



# Copyright (C) 2021 ASTRON (Netherlands Institute for Radio Astronomy)
# SPDX-License-Identifier: GPL-3.0-or-later

import glob
import matplotlib.pyplot as plt
import numpy as np
from astropy.io import fits

import matplotlib

matplotlib.use("Agg")

"""
Plotting utilities that can be used to visualize parts of the OSKAR demos.
"""


def plot_fits_panel(ax, image, title, cmap):
    """
    Plot a single fits panel

    Parameters
    ----------
    ax : matplotlib.axis
        Axis to which image should be added
    image : np.ndarray
        Array containing the image data.
    title : str
        Title for plot
    cmap : matplotlib.colormap
        Colormap to use
    """

    im = ax.imshow(np.squeeze(image), cmap=cmap)
    plt.colorbar(im, format="%.2e")
    plt.tick_params(
        labelcolor="none", top="off", bottom="off", left="off", right="off"
    )
    ax.set_xticks([])
    ax.set_yticks([])
    ax.invert_yaxis()
    ax.set_title(title)
    ax.axis("equal")


def make_fits_plots(title, glob_pattern, out_basename):
    """
    Generates a plot of the fits image with four panels
    (either all Stokes components, or all polarizations)

    Parameters
    ----------
    title : str
        Title for plot.
    glob_pattern : str
        Globbing pattern for (fits) filenames that are to be plotted.
    out_basename : str
        Basename for output images.
    """

    # Load FITS images matching the glob pattern.
    files = glob.glob(glob_pattern)
    files.sort()  # Must be sorted!
    images = []
    for file in files:
        images.append(fits.getdata(file))

    # Set titles and colour maps to use.
    cmap = ""
    titles = []
    if "_AUTO_POWER" in glob_pattern:
        cmap = "Blues_r"
        titles = ["Stokes I", "Stokes Q", "Stokes U", "Stokes V"]
    elif "_AMP" in glob_pattern:
        cmap = "CMRmap"
        titles = ["XX", "XY", "YX", "YY"]

    # Sub-plots, one for each Stokes parameter.
    fig = plt.figure(figsize=(8, 6))
    ax = fig.add_subplot(221, frameon=False)
    plot_fits_panel(ax, images[0], titles[0], cmap)
    ax = fig.add_subplot(222, frameon=False)
    plot_fits_panel(ax, images[1], titles[1], cmap)
    ax = fig.add_subplot(223, frameon=False)
    plot_fits_panel(ax, images[2], titles[2], cmap)
    ax = fig.add_subplot(224, frameon=False)
    plot_fits_panel(ax, images[3], titles[3], cmap)

    # Add main title.
    title = title.replace("_", " ")  # Replace underscores with spaces.
    fig.suptitle(title)
    fig.tight_layout()
    fig.subplots_adjust(top=0.88)

    # Save and close.
    plt.savefig(f"{out_basename}.png")
    plt.close("all")



from astropy.coordinates import AltAz, EarthLocation, ITRS, SkyCoord
from astropy.time import Time

import astropy.units as u
import casacore.tables as ct
import everybeam as eb
import numpy as np


def radec_to_xyz(ra, dec, time):
    """Convert ra and dec coordinates to ITRS coordinates for LOFAR observations.

    Args:
        ra (astropy Quantity): right ascension
        dec (astropy Quantity): declination
        time (float): MJD time in seconds
    Returns:
        pointing_xyz (ndarray): NumPy array containing the X, Y and Z coordinates
    """
    obstime = Time(time / 3600 / 24, scale="utc", format="mjd")
    loc_LOFAR = EarthLocation(
        lon=0.11990128407256424 * u.rad,
        lat=0.9203091252660295 * u.rad,
        height=6364618.852935438 * u.m,
    )

    dir_pointing = SkyCoord(ra, dec)
    dir_pointing_altaz = dir_pointing.transform_to(
        AltAz(obstime=obstime, location=loc_LOFAR)
    )
    dir_pointing_xyz = dir_pointing_altaz.transform_to(ITRS)

    pointing_xyz = np.asarray(
        [dir_pointing_xyz.x, dir_pointing_xyz.y, dir_pointing_xyz.z]
    )
    return pointing_xyz


# Set path to LOFAR LBA MS and load telescope
ms_path = "/path/to/my.ms"

telescope = eb.load_telescope(ms_path)
assert type(telescope) == eb.LOFAR

# Time slots at which to evaluate the beam response.
ms_times = ct.taql("SELECT TIME FROM {ms:s}".format(ms=ms_path))
times = ms_times.getcol("TIME")[0:5]

# Frequencies at which to evaluate the beam response.
ms_freqs = ct.taql(
    "SELECT CHAN_FREQ FROM {ms:s}::SPECTRAL_WINDOW".format(ms=ms_path)
)
freqs = ms_freqs.getcol("CHAN_FREQ").squeeze()

# Obtain the reference direction from the Measurement Set.
ms_dirs = ct.taql("SELECT REFERENCE_DIR FROM {ms:s}::FIELD".format(ms=ms_path))
ra_ref, dec_ref = ms_dirs.getcol("REFERENCE_DIR").squeeze()

# Change these to the RA and DEC of interest, in units of radians.
ra, dec = 1.34, 1.56

# Here we obtain an array of (x, y, z) coordinates for each time slot in the Measurement Set.
# ITRF coordinates of the reference direction.
reference_xyz = list(
    zip(*radec_to_xyz(ra_ref * u.rad, dec_ref * u.rad, times))
)
# ITRF coordinates of the phase centre to correct the array factor for.
phase_xyz = list(zip(*radec_to_xyz(ra * u.rad, dec * u.rad, times)))
# Get teh array factor for station 0
station_id = 0
array_factor = telescope.array_factor(
    times[0], station_id, freqs[0], phase_xyz[0], reference_xyz[0]
)

print(array_factor)



#!/usr/bin/env python3
# Copyright (C) 2021 ASTRON (Netherlands Institute for Radio Astronomy)
# SPDX-License-Identifier: GPL-3.0-or-later

import everybeam as eb
import numpy as np

"""
Demo script illustrating how the EveryBeam Python
bindings can be used to compute the beam response on
a grid for LOFAR observations.

Script requires an LOFAR Measurement Set in order to run.
A minimal LOFAR MSet can be downloaded here:
https://support.astron.nl/software/ci_data/EveryBeam/lba.MS.tar.bz2
"""

# Set path to LOFAR LBA MS and load telescope
ms_path = os.path.join(os.environ["DATA_DIR"], "LOFAR_LBA_MOCK.ms")

telescope = eb.load_telescope(ms_path)
assert type(telescope) == eb.LOFAR

# Set time and freq for which beam should be evaluated
time = 4.92183348e09
freq = 57884216.30859375
station_id = 23

# Specify the settings of the grid (/image)
gs = eb.GridSettings()
gs.width = gs.height = 4
gs.ra = -1.44194878
gs.dec = 0.85078091
gs.dl = gs.dm = 0.5 * np.pi / 180.0
gs.l_shift = gs.m_shift = 0.0

# Get the gridded response for all stations at once
grid_response_all = telescope.gridded_response(gs, time, freq)

# Check whether the returned numpy array has the expected shape
assert grid_response_all.shape == (
    telescope.nr_stations,
    gs.width,
    gs.height,
    2,
    2,
)

# Get the response for a specific station (station_id)
grid_response = telescope.gridded_response(gs, time, freq, station_id)

# Returned response should of course match corresponding entry
# in "grid_response_all"
np.testing.assert_allclose(
    grid_response, grid_response_all[station_id, ...], rtol=1e-6
)

# Also, at the center pixel (2,2) the retrieved response should be reproduced
# by a call to station_response at the phase centre ra and dec
station_response = telescope.station_response(
    time, station_id, freq, gs.ra, gs.dec
)
np.testing.assert_allclose(
    grid_response[2, 2, ...], station_response, rtol=1e-6
)



#!/usr/bin/env python3
from idg.h5parmwriter import H5ParmWriter
import numpy as np
from datetime import datetime

"""
Script for generating the MOCK_H5PARM.h5 file, used for unittesting the
H5ParmATerm implementation.
Please note that running this script requires the python idg module
to be on your PYTHONPATH.
"""

## INPUT
fname = "MOCK_H5PARM.h5"
solsetname = "coefficients000"

antenna_names = np.array(["Antenna0", "Antenna1"])
antenna_positions = np.array([[0.0, 1.0, 2.0], [3, 4, 5]])

nr_stations = antenna_names.size
# Amplitude --> Second order polynomial
nr_parameters_ampl = 6
time_array_ampl = np.linspace(0, 10, num=3)

# Phase --> First order polynomial
nr_parameters_phase = 3
time_array_phase = np.linspace(0, 10, num=11)

image_size = 1.0
subgrid_size = 32
## END INPUT


def init_h5parm_solution_table(
    h5parm_object,
    soltab_type,
    axes_info,
    antenna_names,
    time_array,
    image_size,
    subgrid_size,
    basisfunction_type="lagrange",
    history="",
):
    soltab_info = {
        "amplitude": "amplitude_coefficients",
        "phase": "phase_coefficients",
    }

    assert soltab_type in soltab_info.keys()
    soltab_name = soltab_info[soltab_type]

    h5parm_object.create_solution_table(
        soltab_name,
        soltab_type,
        axes_info,
        dtype=np.float_,
        history=f'CREATED at {datetime.today().strftime("%Y/%m/%d")}; {history}',
    )

    # Set info for the "ant" axis
    h5parm_object.create_axis_meta_data(
        soltab_name, "ant", meta_data=antenna_names
    )

    # Set info for the "dir" axis
    h5parm_object.create_axis_meta_data(
        soltab_name,
        "dir",
        attributes={
            "basisfunction_type": basisfunction_type,
            "image_size": image_size,
            "subgrid_size": subgrid_size,
        },
    )

    # Set info for the "time" axis
    h5parm_object.create_axis_meta_data(
        soltab_name, "time", meta_data=time_array
    )
    return h5parm_object


axes_labels = ["ant", "time", "dir"]
axes_data_amplitude = dict(
    zip(
        axes_labels,
        (nr_stations, time_array_ampl.size, nr_parameters_ampl),
    )
)

axes_data_phase = dict(
    zip(
        axes_labels,
        (
            nr_stations,
            time_array_phase.size,
            nr_parameters_phase,
        ),
    )
)

h5writer = H5ParmWriter(fname, solution_set_name=solsetname, overwrite=True)


h5writer.add_antennas(antenna_names, antenna_positions)

h5writer = init_h5parm_solution_table(
    h5writer,
    "amplitude",
    axes_data_amplitude,
    antenna_names,
    time_array_ampl,
    image_size,
    subgrid_size,
)

h5writer = init_h5parm_solution_table(
    h5writer,
    "phase",
    axes_data_phase,
    antenna_names,
    time_array_phase,
    image_size,
    subgrid_size,
)


amplitude_coeffs = np.array([[1.0, 2.0, 3, 4, 5, 6], [7, 8, 9, 10, 11, 12]])
for i, tampl in enumerate(time_array_ampl):
    amplitude_coeffs_tmp = amplitude_coeffs * (i + 1)
    amplitude_coeffs_tmp = amplitude_coeffs_tmp.reshape(
        nr_stations, 1, nr_parameters_ampl
    )
    offset_amplitude = (0, i, 0)
    h5writer.fill_solution_table(
        "amplitude_coefficients", amplitude_coeffs_tmp, offset_amplitude
    )

phase_coeffs = np.array([[1, 2, 3], [4, 5, 6]])
for i, tphase in enumerate(time_array_phase):
    phase_coeffs_tmp = phase_coeffs * (i + 1)
    phase_coeffs_tmp = phase_coeffs_tmp.reshape(
        nr_stations, 1, nr_parameters_phase
    )
    offset_phase = (0, i, 0)
    h5writer.fill_solution_table(
        "phase_coefficients", phase_coeffs_tmp, offset_phase
    )



#!/usr/bin/env python3
# Copyright (C) 2020 ASTRON (Netherlands Institute for Radio Astronomy)
# SPDX-License-Identifier: GPL-3.0-or-later
"""
Generates all-sky zenith-centred beam patterns for SKALA-4 and EDA-2 antennas.
"""

import copy
import os
import subprocess

# pylint: disable=wrong-import-position
from oskar_utils import create_settings, get_telescope_settings


def main(npixels):
    """Main function."""
    # Name of the application to run, and a settings file for it.
    app1 = "oskar_sim_interferometer"
    app2 = "oskar_sim_beam_pattern"
    settings_path = "_temp_settings.ini"

    data_dir = os.environ["DATA_DIR"]

    # Define some basic observation parameters.
    ra0_deg = 0.0
    dec0_deg = -27.0
    length_sec = 60  # 3600.0*6
    # Define frequencies of interest (in MHz).
    freqs = [50]

    telescope_settings = get_telescope_settings(
        data_dir,
        ra0_deg,
        dec0_deg,
        length_sec,
        num_time_steps=60,
        swap_xy=False,
    )

    # Copy the base settings dictionary.
    current_settings = copy.deepcopy(telescope_settings)

    # Loop over frequencies.
    for freq in freqs:
        # Create the settings file.
        settings = create_settings(app1, settings_path, current_settings)

        # Update output root path and frequency.
        settings["observation/start_frequency_hz"] = 1e6 * freq
        settings["interferometer/ms_filename"] = "skalowmini-coef.MS"

        # Run the app with the settings file.
        subprocess.call([app1, settings_path])

        # Create the settings file.
        settings = create_settings(app2, settings_path, current_settings)

        # Update frequency.
        root_path = "basefunctions" + f"_{int(freq):03}_MHz"
        settings["beam_pattern/root_path"] = root_path
        settings["observation/start_frequency_hz"] = 1e6 * freq
        settings["beam_pattern/coordinate_frame"] = "Horizon"
        settings["beam_pattern/beam_image/size"] = npixels
        settings["beam_pattern/station_outputs/fits_image/amp"] = True
        settings["beam_pattern/station_outputs/fits_image/phase"] = True
        settings["beam_pattern/station_outputs/fits_image/auto_power_real"] = (
            False
        )

        # Run the app with the settings file.
        subprocess.call([app2, settings_path])


if __name__ == "__main__":
    main(256)



# Copyright (C) 2022 ASTRON (Netherlands Institute for Radio Astronomy)
# SPDX-License-Identifier: GPL-3.0-or-later

import everybeam as eb
import pytest
import os
from subprocess import check_call
import numpy as np
import warnings


EVERYBEAM_BASE_URL = "https://support.astron.nl/software/ci_data/EveryBeam/"
DATADIR = os.environ["DATA_DIR"]
RASCIL_FITS = "test_primary_beam_RADEC_MID_512.fits"


@pytest.fixture(autouse=True)
def download_rascil_fits():
    if not os.path.isfile(os.path.join(DATADIR, RASCIL_FITS)):
        print(os.getcwd())
        wget = f"wget -q {os.path.join(EVERYBEAM_BASE_URL, RASCIL_FITS)} -O {os.path.join(DATADIR, RASCIL_FITS)}"
        check_call(wget.split())


@pytest.fixture
def skamid_setup():
    gs = eb.GridSettings()
    gs.width = gs.height = 512
    gs.ra = 0.0
    gs.dec = -0.785398
    gs.dl = gs.dm = 8.0 / gs.width * np.pi / 180.0
    gs.l_shift = gs.m_shift = 0.0

    return {
        "filename": "SKA_MID_MOCK.ms",
        # Time not relevant for the analytical SKA-MID response model
        "time": 0.0,
        "freq": 1.3e9,
        "coordinate_system": gs,
        "station_id": 0,
        "response_model": "skamid_analytical",
    }


@pytest.mark.parametrize("settings", [pytest.lazy_fixture("skamid_setup")])
def test_skamid_analytical(settings):
    """
    Compare SKA-mid (analytical) gridded response against results obtained with
    RASCIL
    """

    try:
        from astropy.io import fits
    except:
        warnings.warn(
            UserWarning(
                "Could not import astropy, so fits image checks are skipped."
            )
        )
        return

    ms_path = os.path.join(DATADIR, settings["filename"])
    rascil_fits = os.path.join(DATADIR, RASCIL_FITS)

    telescope = eb.load_telescope(
        ms_path, element_response_model=settings["response_model"]
    )
    assert type(telescope) is eb.SkaMid

    grid_response = telescope.gridded_response(
        settings["coordinate_system"],
        settings["time"],
        settings["freq"],
        settings["station_id"],
    )

    rascil_response = fits.open(rascil_fits)[0].data
    # In rascil, the voltage pattern is multiplied with its complex conjugate - even though the voltage
    # pattern is real-valued.
    dimage = (
        grid_response[..., 0, 0].flatten()
        * np.conj(grid_response[..., 0, 0].flatten())
        - rascil_response.flatten()
    )
    rms = np.sqrt(dimage.dot(dimage) / dimage.size)
    assert rms < 1e-5



#!/usr/bin/env python3
# Copyright (C) 2020 ASTRON (Netherlands Institute for Radio Astronomy)
# SPDX-License-Identifier: GPL-3.0-or-later

import matplotlib

matplotlib.use("tkagg")
from matplotlib import pyplot as plt
from scipy.io import loadmat
import lobes
import numpy as np

# create a vectorized element_response_lba function from the scalar function in lobes module
# results are stacked in one array
element_response_lba = lambda *args: np.stack(
    np.frompyfunc(lobes.element_response_lba, 3, 1)(*args)
)

if "simdata" not in globals():
    simdata = loadmat("LBA_CS302_fine")

if "coords" not in globals():
    coords = loadmat("CS302_coords")

P = coords["LBA"]["P"][0, 0][:, 0]
Q = coords["LBA"]["Q"][0, 0][:, 0]
positions = np.array([P, Q])

plt.figure()
for i, (p, q) in enumerate(zip(P, Q)):
    plt.plot(p, q, "x")
    plt.text(p, q, str(i))

theta_idx = -2

freq_idx = 0

theta_deg = simdata["Theta"][0, theta_idx]
phi_deg = simdata["Phi"][0]

E1 = simdata["Vdiff_pol1"].copy()
E2 = simdata["Vdiff_pol2"].copy()

E_norm = np.sqrt(
    np.mean(
        abs(E1[:, :, 0, :, :] ** 2 + abs(E2[:, :, 0, :, :]) ** 2),
        axis=(0, 1, 2),
    )
)

E1 /= E_norm[np.newaxis, np.newaxis, np.newaxis, np.newaxis, :]
E2 /= E_norm[np.newaxis, np.newaxis, np.newaxis, np.newaxis, :]

for el in [0, 1]:
    plt.figure()
    plt.plot(phi_deg, abs(E1[el, :, theta_idx, :, freq_idx].T))
    plt.plot(phi_deg, abs(E2[el, :, theta_idx, :, freq_idx].T))

    plt.legend(
        [
            "X dipole, Etheta",
            "X dipole, Ephi",
            "Y dipole, Etheta",
            "Y dipole, Ephi",
        ]
    )
    plt.title(
        "element = {element}, theta = {theta:f}".format(
            element=el, theta=theta_deg
        )
    )
    plt.xlabel("phi(deg)")
    plt.ylabel("E (V)")

theta = theta_deg / 180.0 * np.pi
# phi = np.mod((simdata['Phi'][0, :] - 45) / 180.0 * np.pi, 2*np.pi)

phi_deg = np.linspace(0, 360, 361)
phi = (phi_deg - 45) / 180.0 * np.pi

freq = simdata["Freq"][0, freq_idx]

r = element_response_lba(freq, theta, phi)

plt.figure()
plt.plot(phi_deg, abs(r[:, 0, 0]))
plt.plot(phi_deg, abs(r[:, 1, 0]))
plt.plot(phi_deg, abs(r[:, 0, 1]))
plt.plot(phi_deg, abs(r[:, 1, 1]))
plt.title("Hamaker, theta = {theta:f}".format(element=el, theta=theta_deg))
plt.xlabel("phi(deg)")
plt.ylabel("E (V)")

plt.show()



#!/usr/bin/env python3
# Copyright (C) 2020 ASTRON (Netherlands Institute for Radio Astronomy)
# SPDX-License-Identifier: GPL-3.0-or-later
"""
Generates all-sky zenith-centred beam patterns for SKALA-4 and EDA-2 antennas.
"""

import copy
import os.path
import subprocess

import matplotlib

matplotlib.use("Agg")
# pylint: disable=wrong-import-position
import oskar
from oskar_utils import get_telescope_settings


def main():
    """
    Main function for running the OSKAR simulation.
    """

    # Name of the application to run, and a settings file for it.
    app1 = "oskar_sim_interferometer"
    settings_path = "_temp_settings.ini"
    data_dir = os.environ["DATA_DIR"]
    src_dir = os.path.dirname(__file__)

    # Define some basic observation parameters.
    ra0_deg = 20.0
    dec0_deg = -30.0
    length_sec = 3600.0 * 8
    freqs = [50]

    # Define telescope settings dictionary
    telescope_settings = get_telescope_settings(
        data_dir,
        ra0_deg,
        dec0_deg,
        length_sec,
        num_time_steps=600,
        swap_xy=True,
    )

    # Copy the base settings dictionary
    current_settings = copy.deepcopy(telescope_settings)

    # Loop over frequencies.
    for freq in freqs:
        # Create the settings file.
        open(settings_path, "w").close()
        settings = oskar.SettingsTree(app1, settings_path)
        settings.from_dict(current_settings)

        # Update output root path and frequency.
        settings["observation/start_frequency_hz"] = 1e6 * freq
        settings["interferometer/ms_filename"] = "testdata.ms"
        settings["sky/oskar_sky_model/file"] = os.path.join(src_dir, "sky.osm")

        # Run the app with the settings file.
        subprocess.call([app1, settings_path])


if __name__ == "__main__":
    main()

    # Add beam info to testdata.ms
    data_dir = os.environ["DATA_DIR"]
    subprocess.check_call(
        [
            "add_beaminfo.py",
            "testdata.ms",
            os.path.join(data_dir, "skalowmini-coef.tm"),
        ]
    )



#!/usr/bin/env python3

"""
Script to create and publish a Debian repository on a Debian package repository server.
"""

import argparse
import logging
import sys
from aptly_api import Client
from aptly_api.base import AptlyAPIException

logging.basicConfig(level=logging.INFO, format="%(levelname)s: %(message)s")
logger = logging.getLogger()


def parse_arguments():
    """
    Parse command-line arguments.
    :return: argparse.Namespace
    """
    # Do not add help automatically to prevent that two sections of optional arguments appear
    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                     description="Create and publish a Debian repository "
                                                 "on a Debian package repository server.",
                                     add_help=False)
    required = parser.add_argument_group("required arguments")
    optional = parser.add_argument_group("optional arguments")
    required.add_argument("reponame",
                          help="name of the repository to create")
    required.add_argument("-a", "--architectures", nargs="+", required=True,
                          help="white-space separated list of supported architectures (e.g. amd64)")
    required.add_argument("-d", "--distribution", required=True,
                          help="default distribution when publishing (e.g. bionic)")
    optional.add_argument("-c", "--component",
                          help="default component when publishing (e.g. main)")
    # Add help manually to optional group.
    optional.add_argument("-h", "--help", action="help",
                          help="show this help message and exit")
    optional.add_argument("-p", "--prefix", default=None,
                          help="publishing prefix (default: <reponame>)")
    optional.add_argument("-r", "--rollback", action="store_true",
                          help="roll back all changes upon failure")
    optional.add_argument("-s", "--skip-signing", action="store_true",
                          help="skip signing when publishing repository (not recommended)")
    optional.add_argument("-u", "--url", default="http://packages.astron.nl:8080",
                          help="URL of repository server (default: %(default)s)")
    optional.add_argument("--gpg-key",
                          help="GPG key ID to use when signing the release")
    optional.add_argument("--gpg-passphrase",
                          help="GPG passphrase to unlock private key (possibly insecure)")
    args = parser.parse_args()
    args.prefix = args.reponame if args.prefix is None else args.prefix
    return args


def main(args):
    """
    Main function. It creates a new repository on the repository server, based on the command line
    arguments that were supplied.
    :param args: Arguments as parsed by `parse_arguments()`.
    :return: Exit status: 0: success; 1: error
    """
    # Open connection to the server
    aptly = Client(args.url)

    # Create repository on the server
    try:
        result = aptly.repos.create(args.reponame,
                                    default_component=args.component,
                                    default_distribution=args.distribution)
        logger.info("Created repository %s: %s", args.reponame, result)
    except AptlyAPIException as excp:
        logger.error("Failed to create repository %s: %s", args.reponame, excp)
        return 1

    # Publish the repository on the server
    try:
        if args.skip_signing:
            result = aptly.publish.publish(sources=[{'Name': args.reponame}],
                                           architectures=args.architectures,
                                           distribution=args.distribution,
                                           prefix=args.prefix,
                                           sign_skip=True)
        else:
            result = aptly.publish.publish(sources=[{'Name': args.reponame}],
                                           architectures=args.architectures,
                                           distribution=args.distribution,
                                           prefix=args.prefix,
                                           sign_gpgkey=args.gpg_key,
                                           sign_passphrase=args.gpg_passphrase,
                                           sign_batch=True)
        logger.info("Published repository %s: %s", args.reponame, result)
    except AptlyAPIException as excp:
        logger.error("Failed to publish repository %s: %s", args.reponame, excp)
        # Roll back any changes if requested by the user
        if args.rollback:
            try:
                aptly.repos.delete(args.reponame)
                logger.info("Rolled back changes")
            except AptlyAPIException as excp:
                logger.error("Failed to rollback changes: %s", excp)
        return 1

    return 0


if __name__ == "__main__":
    sys.exit(main(parse_arguments()))



#!/usr/bin/env python3
# Copyright (C) 2021 ASTRON (Netherlands Institute for Radio Astronomy)
# SPDX-License-Identifier: GPL-3.0-or-later

import numpy as np
from astropy.time import Time, TimeDelta
from astropy.io import fits
import oskar
import os

"""
Common functions used in the demo project.
"""


def check_tolerance(tolerance, A, B):
    """
    Checks that the maximum (absolute) difference between two matrices are no bigger than the tolerance.

    Parameters
    ----------
    tolerance : float
        (absolute) tolerance
    A : np.ndarray
        Matrix A
    B : np.ndarray
        Matrix B
    """
    if tolerance:
        np.testing.assert_allclose(
            A,
            B,
            atol=tolerance,
            equal_nan=True,
            err_msg=f"Difference between OSKAR and EveryBeam spherical wave model exceeds tolerance of {tolerance}",
        )


def create_settings(app, settings_path, current_settings):
    """
    Create settings file

    app :
        OSKAR application
    settings_path : str
        Path to settings file
    current_settings : dict
        Dictionary with settings
    """
    open(settings_path, "w").close()
    settings = oskar.SettingsTree(app, settings_path)
    settings.from_dict(current_settings)
    return settings


def get_start_time(ra0_deg, length_sec, optimal_time="2000-01-01 12:00:00"):
    """
    Returns optimal start time for field RA and observation length

    Parameters
    ----------
    ra0_deg : float
        Right Ascension (in degrees)
    length_sec : float
        Observation in length
    optimal_time : str, optional
        Optimal time, by default "2000-01-01 12:00:00"

    Returns
    -------
    float
        Start time
    """
    t = Time(optimal_time, scale="utc", location=("116.764d", "0d"))
    dt_hours = (24.0 - t.sidereal_time("apparent").hour) / 1.0027379
    dt_hours += ra0_deg / 15.0
    start = t + TimeDelta(dt_hours * 3600.0 - length_sec / 2.0, format="sec")
    return start.value


def get_telescope_settings(
    data_dir, ra0_deg, dec0_deg, length_sec, num_time_steps, swap_xy
):
    """
    Generate a dict with telescope settings.

    Parameters
    ----------
    data_dir : str
        Path to data directory
    ra0_deg : float
        Right ascension of observation
    dec0_deg : float
        Declination of observation
    length_sec : float
        Observation duration in seconds.
    num_time_steps : int
        Number of time steps
    swap_xy : bool
        Swap x- and y-axis in element coordinate system?

    Returns
    -------
    dict
        Dictionary with telescope settings.
    """
    # Define base settings dictionary.
    return {
        "observation": {
            "phase_centre_ra_deg": ra0_deg,
            "phase_centre_dec_deg": dec0_deg,
            "start_time_utc": get_start_time(
                ra0_deg, length_sec, "2000-01-01 12:00:00"
            ),
            "length": length_sec,
            "num_time_steps": num_time_steps,
        },
        "telescope": {
            "normalise_beams_at_phase_centre": False,
            "aperture_array/element_pattern/normalise": False,
        },
        "telescope/input_directory": os.path.join(
            data_dir, "skalowmini-coef.tm"
        ),
        "telescope/aperture_array/element_pattern/enable_numerical": True,
        "telescope/aperture_array/element_pattern/swap_xy": swap_xy,
        "telescope/aperture_array/array_pattern/enable": True,
        "telescope/aperture_array/array_pattern/normalise": True,
    }


def generate_oskar_csv(
    basefunction_idx, freqs, output_dir, element_id, em_idx=0
):
    """
    Write OSKAR results to csv file.
    """

    l_max = int(np.sqrt(basefunction_idx + 1))

    # Parse all freqs
    for freq in freqs:
        A = np.zeros((l_max, 2 * l_max + 1, 2, 2), dtype=np.complex128)
        data = np.zeros((l_max * (l_max + 2), 4), dtype=np.complex128)
        data[basefunction_idx, em_idx] = 1.0

        i = 0
        for l in range(l_max):
            n = (l + 1) * 2 + 1
            A[l, :n, :] = data[i : i + n, :].reshape(n, 2, 2)
            i += n

        for i, pol in enumerate(("x", "y")):
            for j, em in enumerate(("e", "m")):
                for reim, s in zip(("re", "im"), (1.0, 1.0j)):
                    filename = f"element_pattern_spherical_wave_{pol}_t{em}_{reim}_{element_id}_{freq}.txt"
                    alpha = np.real(A[:, :, i, j] / s)

                    np.savetxt(
                        os.path.join(output_dir, filename),
                        alpha,
                        delimiter=", ",
                    )


def read_oskar_beams(root_name):
    """
    Read oskar beam and store into 4-d numpy array.

    Parameters
    ----------
    root_name : str
        Root name of beam file.

    Returns
    -------
    np.array
        (N, N, 2, 2) array, where N number of image pixels
    """
    A = None
    for i, pol1 in enumerate(["X", "Y"]):
        for j, pol2 in enumerate(["X", "Y"]):
            filename_amp = f"{root_name}_MHz_S0000_TIME_SEP_CHAN_SEP_AMP_{pol1}{pol2}.fits"
            filename_phase = f"{root_name}_MHz_S0000_TIME_SEP_CHAN_SEP_PHASE_{pol1}{pol2}.fits"
            d = (
                fits.getdata(filename_amp)
                * np.exp(1j * fits.getdata(filename_phase))
            )[0, 0, ...].T
            if A is None:
                N = d.shape[-1]
                A = np.zeros((N, N, 2, 2), dtype=np.complex128)

            # One axis in the OSKAR fits files runs in opposite direction compared to the
            # numpy arrays made by make_element_response_image.cpp
            # we flip that axis here to make them the same
            A[..., i, j] = d[::-1, :]

        N = A.shape[0]

    for i in range(N):
        x = 2.0 * i / (N - 1) - 1.0
        for j in range(N):
            y = 2.0 * j / (N - 1) - 1.0
            phi = np.arctan2(y, x)

            e_theta = np.array([[np.cos(phi)], [np.sin(phi)]])
            e_phi = np.array([[-np.sin(phi)], [np.cos(phi)]])

            # This matrix applies the transformation defined in
            # https://github.com/OxfordSKA/OSKAR/blob/master/oskar/convert/define_convert_ludwig3_to_theta_phi_components.h
            T = np.concatenate((e_theta, e_phi), axis=1)

            # Transformation is applied to the right side of A
            # so the rows a of A are tranformed
            A[i, j, ...] = np.dot(A[i, j, ...], T)
    return A



#!/usr/bin/env python3

"""
Script to update a published Debian repository on a Debian package repository server.
"""

import argparse
import logging
import sys
from aptly_api import Client
from aptly_api.base import AptlyAPIException

logging.basicConfig(level=logging.INFO, format='%(levelname)s: %(message)s')
logger = logging.getLogger()

def parse_arguments():
    """
    Parse command-line arguments.
    :return: argparse.Namespace
    """
    # Do not add help automatically to prevent that two sections of optional arguments appear
    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                     description="Update a published Debian repository "
                                                 "on a Debian package repository server.",
                                     add_help=False)
    required = parser.add_argument_group("required arguments")
    optional = parser.add_argument_group("optional arguments")
    required.add_argument("reponame",
                          help="name of the repository to update")
    required.add_argument("files", nargs="+",
                          help="white-space separated list of files to upload and publish")
    required.add_argument("-d", "--distribution", required=True,
                          help="name of the distribution (e.g. bionic)")
    # Add help manually to optional group.
    optional.add_argument("-h", "--help", action="help",
                          help="show this help message and exit")
    optional.add_argument("-p", "--prefix", default=None,
                          help="publishing prefix (default: <reponame>)")
    optional.add_argument("-s", "--skip-signing", action="store_true",
                          help="skip signing when publishing repository (not recommended)")
    optional.add_argument("-u", "--url", default="http://packages.astron.nl:8080",
                          help="URL of repository server (default: %(default)s)")
    optional.add_argument("--cleanup", action="store_true",
                          help="cleanup upload directory on the server")
    optional.add_argument("--gpg-key",
                          help="GPG key ID to use when signing the release")
    optional.add_argument("--gpg-passphrase",
                          help="GPG passphrase to unlock private key (possibly insecure)")
    optional.add_argument("--upload-dir", default=None,
                          help="upload directory on the server to temporarily store files "
                               "(default: <reponame>)")
    args = parser.parse_args()
    args.prefix = args.reponame if args.prefix is None else args.prefix
    args.upload_dir = args.reponame if args.upload_dir is None else args.upload_dir
    return args


def main(args):
    """
    Main function. It creates a new repository on the repository server, based on the command line
    arguments that were supplied.
    :param args: Arguments as parsed by `parse_arguments()`.
    """
    # Open connection to the server
    aptly = Client(args.url)
    status = 0

    try:
        # Upload file(s) to the server
        result = aptly.files.upload(args.upload_dir,
                                    *args.files)
        logger.info("Uploaded file(s): %s", " ".join(result))

        # Add uploaded files to the local repository; this will automatically remove them from the
        # upload directory
        result = aptly.repos.add_uploaded_file(reponame=args.reponame,
                                               force_replace=True,
                                               dir=args.upload_dir)
        added_files = result.report['Added']
        removed_files = result.report['Removed']

        if added_files or removed_files:
            added = "added: %s" % ", ".join(added_files) if added_files else ""
            removed = "removed: %s" % ", ".join(removed_files) if removed_files else ""
            logger.info("Updated local repository: %s", "; ".join((added, removed)))

        if result.failed_files:
            for warning in result.report['Warnings']:
                logger.warning(warning)
            logger.error("Updating file(s): %s", ", ".join(result.failed_files))
            status = 2

        # Update the published repository with the latest local content
        if args.skip_signing:
            result = aptly.publish.update(distribution=args.distribution,
                                        prefix=args.prefix,
                                        sign_skip=True)
        else:
            result = aptly.publish.update(distribution=args.distribution,
                                        prefix=args.prefix,
                                        sign_gpgkey=args.gpg_key,
                                        sign_passphrase=args.gpg_passphrase)
        logger.info("Published %s: %s", args.reponame, result)

    except AptlyAPIException as excp:
        logger.error(excp)
        status = 1

    finally:
        if args.cleanup:
            aptly.files.delete(args.upload_dir)

    return status


if __name__ == "__main__":
    sys.exit(main(parse_arguments()))



# Copyright (C) 2022 ASTRON (Netherlands Institute for Radio Astronomy)
# SPDX-License-Identifier: GPL-3.0-or-later

from everybeam import ElementResponse
from everybeam import ElementResponseModel
import pytest
import numpy as np


@pytest.mark.parametrize("use_string_names", [False, True])
def test_model_types(use_string_names):
    """
    Create differente ElementResponse types and check that the model is correct.
    """
    if use_string_names:
        hamaker_hba = ElementResponse.create("hamaker", "HBA")
        hamaker_lba = ElementResponse.create("hamaker", "LBA")
        lobes = ElementResponse.create("lobes", "CS002LBA")
        lobes_not_found = ElementResponse.create("lobes", "NOT_FOUND_LBA")
        oskar_dipole = ElementResponse.create("oskardipole")
        oskar_wave = ElementResponse.create("oskarsphericalwave")
    else:
        hamaker_hba = ElementResponse.create(
            ElementResponseModel.hamaker, "HBA"
        )
        hamaker_lba = ElementResponse.create(
            ElementResponseModel.hamaker_lba, "LBA"
        )
        lobes = ElementResponse.create(ElementResponseModel.lobes, "CS002LBA")
        lobes_not_found = ElementResponse.create(
            ElementResponseModel.lobes, "NOT_FOUND_LBA"
        )
        oskar_dipole = ElementResponse.create(
            ElementResponseModel.oskar_dipole
        )
        oskar_wave = ElementResponse.create(
            ElementResponseModel.skala40_spherical
        )

    assert hamaker_hba.model == ElementResponseModel.hamaker
    assert hamaker_lba.model == ElementResponseModel.hamaker
    assert lobes.model == ElementResponseModel.lobes
    assert lobes_not_found.model == ElementResponseModel.hamaker
    assert oskar_dipole.model == ElementResponseModel.oskar_dipole
    assert oskar_wave.model == ElementResponseModel.skala40_spherical


def test_response():
    """
    Test that response() supports both single values and arrays, and that the
    returned values are equal. Also test that the values with and without an
    element_id argument are equal.
    """
    element_response = ElementResponse.create("hamaker", "LBA")
    response_10 = element_response.response(10e6, 0.0, 0.0)
    response_11 = element_response.response(11e6, 0.1, 0.1)
    response_12 = element_response.response(12e6, 0.2, 0.1)
    response_10_array = element_response.response([10e6], [0.0], [0.0])
    response_all = element_response.response(
        [10e6, 11e6, 12e6], [0.0, 0.1, 0.2], [0.0, 0.1, 0.1]
    )

    response_10_id = element_response.response(0, 10e6, 0.0, 0.0)
    response_11_id = element_response.response(0, 11e6, 0.1, 0.1)
    response_12_id = element_response.response(0, 12e6, 0.2, 0.1)
    response_10_array_id = element_response.response([0], [10e6], [0.0], [0.0])
    response_all_id = element_response.response(
        [0, 0, 0], [10e6, 11e6, 12e6], [0.0, 0.1, 0.2], [0.0, 0.1, 0.1]
    )

    assert response_10.shape == (2, 2)
    assert response_11.shape == (2, 2)
    assert response_12.shape == (2, 2)
    assert response_10_array.shape == (1, 2, 2)
    assert response_all.shape == (3, 2, 2)

    assert np.array_equal(response_10_array[0], response_10)
    assert np.array_equal(response_all[0], response_10)
    assert np.array_equal(response_all[1], response_11)
    assert np.array_equal(response_all[2], response_12)

    assert np.array_equal(response_10, response_10_id)
    assert np.array_equal(response_11, response_11_id)
    assert np.array_equal(response_12, response_12_id)
    assert np.array_equal(response_10_array, response_10_array_id)
    assert np.array_equal(response_all, response_all_id)



# Copyright (C) 2020 ASTRON (Netherlands Institute for Radio Astronomy)
# SPDX-License-Identifier: GPL-3.0-or-later

from everybeam import load_telescope, LOFAR, GridSettings, ElementResponseModel
import pytest
import os
import numpy as np

DATADIR = os.environ["DATA_DIR"]


@pytest.fixture
def lba_setup():
    gs = GridSettings()
    gs.width = gs.height = 4
    gs.ra = -1.44194878
    gs.dec = 0.85078091
    gs.dl = gs.dm = 0.5 * np.pi / 180.0
    gs.l_shift = gs.m_shift = 0.0

    return {
        "filename": "LOFAR_LBA_MOCK.ms",
        "time": 4.92183348e09,
        # Frequency of channel 4
        "freq": 57884216.30859375,
        "coordinate_system": gs,
        # Reference cpp solution for station 31, see cpp/test/tlofar_lba.test_hamaker
        "station_id": 31,
        "freq0": 5.78125e07,
        "direction": np.array([0.667806, -0.0770635, 0.740335]),
        "preapplied_beam_dir": np.array([0.655743, -0.0670973, 0.751996]),
        "station0": np.array([0.655743, -0.0670973, 0.751996]),
        "station0_name": "CS001LBA",
        "tile0": np.array([0.655743, -0.0670973, 0.751996]),
        "cpp_response": np.array(
            [
                [-0.71383788 + 0.00612506j, -0.4903527 + 0.00171652j],
                [-0.502122 + 0.00821683j, 0.7184408 - 0.00821723j],
            ]
        ),
        # Used for testing the gridded response, note that the used
        # frequency is slighty different from the value in the equivalent cpp test, hence leading to small differences in the reference
        # solution
        "grid_response": {
            "station_id": 31,
            "pixel": (1, 3),
            "response": np.array(
                [
                    [-0.77189475 + 0.00214358j, -0.52923 - 0.00128423j],
                    [-0.5414057 + 0.00531782j, 0.77560395 - 0.00369414j],
                ]
            ),
        },
    }


@pytest.fixture
def hba_setup():
    gs = GridSettings()
    gs.width = gs.height = 4
    gs.ra = 2.15374123
    gs.dec = 0.8415521
    gs.dl = gs.dm = 0.5 * np.pi / 180.0
    gs.l_shift = gs.m_shift = 0.0

    return {
        "filename": "LOFAR_HBA_MOCK.ms",
        "coordinate_system": gs,
        "time": 4929192878.008341,
        "freq": 138476562.5,
        # Reference cpp solution for station 31, see cpp/test/tlofar_lba.test_hamaker
        "station_id": 63,
        "freq0": 138476562.5,
        "direction": np.array([0.42458804, 0.46299569, 0.77804112]),
        "preapplied_beam_dir": np.array([0.408326, 0.527345, 0.745102]),
        "station0": np.array([0.4083262, 0.52734471, 0.74510219]),
        "station0_name": "CS001HBA0",
        "tile0": np.array([0.40832685, 0.52734421, 0.74510219]),
        "cpp_response": np.array(
            [
                [0.03259424 - 0.00023045994j, 0.12204104097 - 0.00091857865j],
                [0.13063535 - 0.0010039175j, -0.02934845 + 0.00023882818j],
            ]
        ),
        # Used for testing the gridded response
        "grid_response": {
            "station_id": 23,
            "pixel": (3, 1),
            "response": np.array(
                [
                    [-0.158755 - 0.000746758j, -0.816172 - 0.00271176j],
                    [-0.863398 - 0.00282507j, 0.0936923 + 0.000109039j],
                ]
            ),
        },
    }


def check_consistency_station_response(telescope, time):
    """
    Check the internal consistency of various routes for computing the station
    response

    Parameters
    ----------
    telescope : everybeam.Telescope
    time : float
        Time MJD in s
    """
    # Checking internal consistency of the different methods
    stations_all = telescope.station_response(time)
    for station_idx in range(stations_all.shape[0]):
        channel_all = telescope.station_response(time, station_idx)
        np.testing.assert_allclose(
            stations_all[station_idx, :, :, :], channel_all
        )
        for channel_idx in range(channel_all.shape[0]):
            channel_single = telescope.station_response(
                time, station_idx, channel_idx
            )
            np.testing.assert_allclose(
                channel_all[channel_idx, :, :], channel_single
            )
            # Check against implementation where frequency is given as input
            frequency = telescope.channel_frequency(channel_idx)
            freq_single = telescope.station_response(
                time, station_idx, frequency
            )
            np.testing.assert_allclose(freq_single, channel_single)


def check_array_factor_vector(telescope, time, direction, station0, tile0):
    frequencies = np.array([100e6, 110e6, 120e6])
    station_indices = np.array([[0, 1, 2, 3], [4, 5, 6, 7]])
    directions = np.stack([direction, direction, direction])
    directions[1, 0] *= -1.0
    directions[2, 1] *= -1.0

    all_array_factors = telescope.array_factor(
        time, station_indices, frequencies, directions, station0, tile0
    )
    assert (
        all_array_factors.shape
        == station_indices.shape
        + frequencies.shape
        + directions.shape[:-1]
        + (2, 2)
    )
    for f in range(frequencies.size):
        for s0 in range(station_indices.shape[0]):
            for s1 in range(station_indices.shape[1]):
                for d in range(directions.shape[0]):
                    single_array_factor = telescope.array_factor(
                        time,
                        station_indices[s0, s1],
                        frequencies[f],
                        directions[d],
                        station0,
                        tile0,
                    )
                    assert single_array_factor.shape == (2, 2)
                    np.testing.assert_allclose(
                        all_array_factors[s0, s1, f, d, :, :],
                        single_array_factor,
                    )


def check_array_factor_all_stations_all_frequencies(
    telescope, time, direction, station0, tile0
):
    all_array_factors = telescope.array_factor(
        time, [], [], direction, station0, tile0
    )
    assert all_array_factors.shape == (
        telescope.nr_stations,
        telescope.nr_channels,
        2,
        2,
    )
    for s in range(telescope.nr_stations):
        for f in range(telescope.nr_channels):
            single_array_factor = telescope.array_factor(
                time,
                s,
                telescope.channel_frequency(f),
                direction,
                station0,
                tile0,
            )
            assert single_array_factor.shape == (2, 2)
            np.testing.assert_allclose(
                all_array_factors[s, f, :, :],
                single_array_factor,
            )


def check_reference_solution(
    telescope,
    time,
    station_id,
    freq,
    direction,
    station0,
    tile0,
    reference_solution,
):
    """
    Check computed response against provided reference solution.
    """
    response = telescope.station_response(
        time,
        station_id,
        freq,
        direction,
        station0,
        tile0,
    )
    np.testing.assert_allclose(response, reference_solution, atol=1e-6)

    # Check that the same response is obtained by multiplying the array factor with the
    # element response
    array_factor = telescope.array_factor(
        time, station_id, freq, direction, station0, tile0
    )
    element_response = telescope.element_response(
        time, station_id, freq, direction
    )
    np.testing.assert_allclose(
        np.matmul(array_factor, element_response), response, atol=1e-6
    )


def test_coordinate_system():
    width_height = 4
    ra_dec = -1.5
    dl_dm = 0.2
    lm_shift = 0.01

    gs = GridSettings()
    gs.width = gs.height = width_height
    gs.ra = gs.dec = ra_dec
    gs.dl = gs.dm = dl_dm
    gs.l_shift = gs.m_shift = lm_shift

    assert gs.width == width_height
    assert gs.height == width_height
    assert gs.ra == ra_dec
    assert gs.dec == ra_dec
    assert gs.dl == dl_dm
    assert gs.dm == dl_dm
    assert gs.l_shift == lm_shift
    assert gs.m_shift == lm_shift


@pytest.mark.parametrize(
    "ref", [pytest.lazy_fixture("lba_setup"), pytest.lazy_fixture("hba_setup")]
)
@pytest.mark.parametrize("differential_beam", [True, False])
def test_lofar(ref, differential_beam):
    ms_path = os.path.join(DATADIR, ref["filename"])
    telescope = load_telescope(
        ms_path, use_differential_beam=differential_beam
    )
    assert type(telescope) is LOFAR
    assert telescope.station_name(0) == ref["station0_name"]
    assert (
        telescope.station_element_response(0).model
        == ElementResponseModel.hamaker
    )

    time = ref["time"]
    freq = ref["freq"]
    direction = (
        ref["preapplied_beam_dir"] if differential_beam else ref["direction"]
    )
    ref_solution = (
        np.eye(2, dtype=np.complex64)
        if differential_beam
        else ref["cpp_response"]
    )

    check_consistency_station_response(telescope, time)
    check_array_factor_vector(
        telescope, time, direction, ref["station0"], ref["tile0"]
    )
    check_array_factor_all_stations_all_frequencies(
        telescope, time, direction, ref["station0"], ref["tile0"]
    )

    check_reference_solution(
        telescope,
        time,
        ref["station_id"],
        freq,
        direction,
        ref["station0"],
        ref["tile0"],
        ref_solution,
    )

    # Array factor in pointing direction should be unity
    array_factor_I = telescope.array_factor(
        time,
        ref["station_id"],
        freq,
        ref["station0"],
        ref["station0"],
    )
    np.testing.assert_allclose(
        array_factor_I, np.eye(2, dtype=np.complex64), rtol=1e-6
    )

    # For equal station and tile direction, check that the same response is
    # obtained via two routes:
    # Response by explicitly specifying the direction
    response_1 = telescope.station_response(
        time,
        ref["station_id"],
        freq,
        ref["direction"],
        ref["station0"],
        ref["station0"],
    )

    # tile0 direction is implicit in station0 direction
    response_2 = telescope.station_response(
        time,
        ref["station_id"],
        freq,
        ref["direction"],
        ref["station0"],
    )
    np.testing.assert_allclose(response_1, response_2)

    # Check that we can reproduce this with array_factor * beam_response
    # Check that the same response is obtained by multiplying the array factor with the
    # element response
    array_factor = telescope.array_factor(
        time, ref["station_id"], freq, ref["direction"], ref["station0"]
    )
    element_response = telescope.element_response(
        time, ref["station_id"], freq, ref["direction"]
    )
    np.testing.assert_allclose(
        np.matmul(array_factor, element_response), response_1, rtol=1e-6
    )

    # Check that is_local flag is propagated
    element_response_local = telescope.element_response(
        time, ref["station_id"], freq, ref["direction"], is_local=True
    )
    assert np.abs(element_response_local[0, 0] - element_response[0, 0]) > 0.01


@pytest.mark.parametrize(
    "ref", [pytest.lazy_fixture("lba_setup"), pytest.lazy_fixture("hba_setup")]
)
def test_lofar_gridded_response(ref):
    """
    Test the gridded response bindings for a LOFAR telescope
    could be merged with test_lofar, but that would result
    in a large and clumsy test
    """

    ms_path = os.path.join(DATADIR, ref["filename"])
    telescope = load_telescope(ms_path)

    grid_response_all = telescope.gridded_response(
        ref["coordinate_system"],
        ref["time"],
        ref["freq"],
    )

    for station_index in range(telescope.nr_stations):
        grid_response = telescope.gridded_response(
            ref["coordinate_system"], ref["time"], ref["freq"], station_index
        )
        np.testing.assert_allclose(
            grid_response, grid_response_all[station_index, ...], rtol=1e-6
        )
        if station_index == ref["station_id"]:
            # Center pixel (2, 2) should be equal to a station_response solution in the phase centre
            station_response = telescope.station_response(
                ref["time"], ref["station_id"], ref["freq"]
            )
            np.testing.assert_allclose(
                grid_response[2, 2, ...], station_response, rtol=1e-6
            )
        if station_index == ref["grid_response"]["station_id"]:
            # Check an off-diagonal entry
            (h_idx, w_idx) = ref["grid_response"]["pixel"]
            np.testing.assert_allclose(
                grid_response[h_idx, w_idx, ...],
                ref["grid_response"]["response"],
                rtol=1e-5,
            )


@pytest.mark.parametrize("ref", [pytest.lazy_fixture("hba_setup")])
def test_lofar_integrated_beam(ref):
    ms_path = os.path.join(DATADIR, ref["filename"])
    telescope = load_telescope(ms_path)

    nbaselines = telescope.nr_stations * (telescope.nr_stations + 1) // 2
    baseline_weights = np.ones(nbaselines, dtype=float)
    undersampling = 2

    # Provide single time
    undersampled_response_0 = telescope.undersampled_response(
        ref["coordinate_system"], ref["time"], ref["freq"], 2, baseline_weights
    )

    # Vector of times
    baseline_weights = np.ones(nbaselines * 2, dtype=float)
    undersampled_response_1 = telescope.undersampled_response(
        ref["coordinate_system"],
        np.array([ref["time"], ref["time"]]),
        ref["freq"],
        undersampling,
        baseline_weights,
    )

    # Response should be the same for vector of identical times and
    # weights
    np.testing.assert_allclose(
        undersampled_response_0, undersampled_response_1, rtol=1e-6
    )

    # Check if a specific value is reproduced at pixel (2,2), see also
    # tlofar_hba::integrated_beam test
    assert abs(undersampled_response_1[2, 2, 0, 0] - 0.0309436) <= 1e-6



#!/usr/bin/env python3
# Copyright (C) 2020 ASTRON (Netherlands Institute for Radio Astronomy)
# SPDX-License-Identifier: GPL-3.0-or-later

import lobes
from scipy.io import loadmat
import numpy as np
import h5py

simdata = loadmat("LBA_CS302_fine")

E1 = simdata[
    "Vdiff_pol1"
].copy()  # X? dipole, shape: N_ant x [Etheta, Ephi] x Ntheta x Nphi x Nfreq
E2 = simdata[
    "Vdiff_pol2"
].copy()  # Y? dipole, shape: N_ant x [Etheta, Ephi] x Ntheta x Nphi x Nfreq


E1 = E1[:, :, 1:91, :-1, :]
E2 = E2[:, :, 1:91, :-1, :]


# with h5py.File('LBA_CS302.hdf5', 'r') as f:
# nms = f["nms"][:]


theta = simdata["Theta"] / 180.0 * np.pi
phi = simdata["Phi"] / 180.0 * np.pi

theta = theta[0, 1:91]
phi = phi[0, 0:-1]

theta_mesh, phi_mesh = np.meshgrid(theta, phi)

theta_mesh_flat = theta_mesh.flatten()
phi_mesh_flat = phi_mesh.flatten()

lbm = lobes.LobesBeamModel("LBA_CS302.hdf5")

r = lbm.eval(theta_mesh_flat, phi_mesh_flat)

Ntheta = len(theta)
Nphi = len(phi)
Nfreq = 4
Nant = 96

rr = r.reshape((Nphi, Ntheta, 2, Nfreq, Nant)).transpose((4, 2, 1, 0, 3))



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

# Make sure that it refers to the shared object file from current build
if "READTHEDOCS" in os.environ:
    sys.path.insert(0, os.path.abspath("../build/python"))
else:
    sys.path.insert(0, os.environ["EVERYBEAM_SO_PATH"])

# -- Project information -----------------------------------------------------

project = "EveryBeam"
copyright = "2022, ASTRON (Netherlands Institute for Radio Astronomy)"
author = "Jakob Maljaars"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx_rtd_theme",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "breathe",
    "myst_parser",
]

# Disable typehints in signatures - doens't seem to take any effect
autodoc_typehints = "none"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

html_static_path = ["_static"]

# Breathe Configuration
# When using CMake, the 'doc' target already sets breathe_projects.
if "READTHEDOCS" in os.environ:
    breathe_projects = {"EveryBeam": "../build/doc/doxygen/xml"}

breathe_default_project = "EveryBeam"



#!/usr/bin/env python3
# Copyright (C) 2020 ASTRON (Netherlands Institute for Radio Astronomy)
# SPDX-License-Identifier: GPL-3.0-or-later

import argparse
import os
import scipy.io as sio
from os.path import dirname, join as pjoin
import numpy as np
import h5py
from tqdm import tqdm

# Parse command-line arguments
parser = argparse.ArgumentParser()
parser.add_argument(
    "input_dir", type=str, help="directory containing coefficients"
)
parser.add_argument(
    "output_file", type=str, help="name of the HDF5 output file"
)
parser.add_argument("-debug", dest="debug", action="store_true")
parser.set_defaults(input_dir=".")
parser.set_defaults(output_file="oskar.h5")
parser.set_defaults(debug=False)
args = parser.parse_args()

# Read the input directory
# This directory should contain subdirectories,
# with names corresponding to frequency.
input_dir = args.input_dir
print("Reading coefficients from: ", input_dir)
dirs = os.listdir(input_dir)
dirs = list(filter(lambda x: x.isdigit(), dirs))
freqs = sorted(map(int, dirs))
print("Parsing frequencies %d-%d MHz" % (int(freqs[0]), int(freqs[-1])))

# Create HDF5 file
output_file = args.output_file
print("Creating HDF5 file: ", output_file)
hf = h5py.File(output_file, "w")

# Debugging
debug = args.debug
if debug:
    freqs = [freqs[0]]
    tqdm = lambda x: x

# Parse all freqs
for freq in tqdm(freqs):
    # Get path to matlab data files
    patha = pjoin(input_dir, str(freq), "alphas_pola.mat")
    pathb = pjoin(input_dir, str(freq), "alphas_polb.mat")

    # Read the matlab data files
    matrixa = sio.loadmat(patha)
    matrixb = sio.loadmat(pathb)

    # Convert to numpy arrays
    alpha_te_a = np.asarray(matrixa["alpha_te"])
    alpha_tm_a = np.asarray(matrixa["alpha_tm"])
    alpha_te_b = np.asarray(matrixb["alpha_te"])
    alpha_tm_b = np.asarray(matrixb["alpha_tm"])

    # Make sure that all four sub-matrices are shaped identically
    assert (
        alpha_te_a.shape
        == alpha_tm_a.shape
        == alpha_te_b.shape
        == alpha_tm_b.shape
    )

    # Get dimensions of input matrix
    # This matrix is lower triangular, e.g.:
    # 32x65 (32x(32*2+1) = 2080 elements),
    # with 1088 (32x(32+2)) nonzeros
    shape = alpha_te_a.shape
    nr_elements = shape[0]
    l_max = shape[1]
    assert shape[2] == ((l_max * 2) + 1)
    if debug:
        print("nr_elements: ", nr_elements)
        print("shape: ", shape)

    # Determine dimensions of output matrix
    # This matrix is rectangular and contains
    # all nonzero elements of the input matrix
    dst_height = l_max + 2
    dst_width = l_max
    if debug:
        print("output height: ", dst_height)
        print("output width: ", dst_width)

    # Fill the output matrix
    # The inner dimension are set as follows:
    # (x_te_re, x_te_im), (x_tm_re, x_tm_im),
    # (y_te_re, y_te_im), (y_tm_re, y_tm_im)
    data = np.zeros(
        (nr_elements, dst_height * dst_width, 4), dtype=np.complex128
    )

    i = 0
    for l in range(l_max):
        n = (l + 1) * 2 + 1
        data[:, i : i + n, 0] = alpha_tm_a[:, l, :n]
        data[:, i : i + n, 1] = alpha_te_a[:, l, :n]
        data[:, i : i + n, 2] = alpha_tm_b[:, l, :n]
        data[:, i : i + n, 3] = alpha_te_b[:, l, :n]
        i += n

    # Sanity check
    assert np.count_nonzero(data[:, :, 0]) == np.count_nonzero(alpha_te_a)
    assert np.count_nonzero(data[:, :, 1]) == np.count_nonzero(alpha_tm_a)
    assert np.count_nonzero(data[:, :, 2]) == np.count_nonzero(alpha_te_b)
    assert np.count_nonzero(data[:, :, 3]) == np.count_nonzero(alpha_tm_b)

    # Add group for current frequency
    hf.create_dataset(str(freq), data.shape, data=data)

# Close HDF5 file
hf.close()

# Done
print("Done!")



#!/usr/bin/env python3
# Copyright (C) 2020 ASTRON (Netherlands Institute for Radio Astronomy)
# SPDX-License-Identifier: GPL-3.0-or-later
"""
Generates all-sky zenith-centred beam patterns for SKALA-4 and EDA-2 antennas.
"""

import copy
import subprocess

from oskar_utils import create_settings, get_telescope_settings
from plot_utils import make_fits_plots


def main(npixels):
    """Main function."""
    # Name of the application to run, and a settings file for it.
    app = "oskar_sim_beam_pattern"
    settings_path = "_temp_settings.ini"

    # Define some basic observation parameters.
    ra0_deg = 0.0
    dec0_deg = -27.0
    length_sec = 1.0
    data_dir = "."  # Not relevant
    num_steps = 1  # Not relevant

    # Define frequencies of interest (in MHz).
    freqs = [50]

    # Define settings dictionary.
    common_settings = get_telescope_settings(
        data_dir, ra0_deg, dec0_deg, length_sec, num_steps, False
    )
    beam_pattern = {
        "beam_pattern": {
            "coordinate_frame": "Horizon",
            "beam_image/size": npixels,
            "station_outputs/fits_image/amp": True,
            "station_outputs/fits_image/phase": True,
            "station_outputs/fits_image/auto_power_real": False,
        }
    }
    # Override and extend telescope settings
    telescope = {
        "telescope/input_directory": "telescope.tm",
        "telescope/aperture_array/element_pattern/enable_numerical": True,
        "telescope/aperture_array/element_pattern/swap_xy": False,
        "telescope/aperture_array/array_pattern/enable": False,
    }
    common_settings = {**common_settings, **beam_pattern, **telescope}

    # Copy the base settings dictionary.
    current_settings = copy.deepcopy(common_settings)

    # Loop over frequencies.
    for freq in freqs:
        # Create the settings file.
        settings = create_settings(app, settings_path, current_settings)

        # Update output root path and frequency.
        root_path = "basefunctions" + f"_{int(freq):03}_MHz"
        settings["beam_pattern/root_path"] = root_path
        settings["observation/start_frequency_hz"] = 1e6 * freq

        # Run the app with the settings file.
        subprocess.call([app, settings_path])

        # [OPTIONALLY] Make plots.
        # title = f"basefunctions @ {freq} MHz"
        # make_fits_plots(title=title,
        # glob_pattern=root_path+'*_AMP*',
        # out_basename=root_path+'_amp')
        # make_fits_plots(title=title,
        # glob_pattern=root_path+'*_AUTO_POWER*',
        # out_basename=root_path+'_stokes')


if __name__ == "__main__":
    main(32)



#!/usr/bin/env python3
# Copyright (C) 2020 ASTRON (Netherlands Institute for Radio Astronomy)
# SPDX-License-Identifier: GPL-3.0-or-later

from everybeam import load_telescope, LOFAR, Options, thetaphi2cart
import matplotlib.pyplot as plt
import numpy as np
import logging

"""
Demo script illustrating how the EveryBeam Python
bindings can be used to generate all-sky images
for the beam response for a LOFAR observation.

Script requires a LOFAR Measurement Set in order to run.
A minimal LOFAR MSet can be downloaded here:
https://support.astron.nl/software/ci_data/EveryBeam/lba.MS.tar.bz2
"""

ms_path = os.path.join(os.environ["DATA_DIR"], "LOFAR_LBA_MOCK.ms")

# Response settings
mode = "station"  # Change to "element" to obtain the element response
time = 4.92183348e09
frequency = 57812500.0
station_id = 20
element_id = 0

# Same station0 direction as in python/test
station0 = np.array([0.655743, -0.0670973, 0.751996])
is_local = True  # use local coords.
rotate = True
response_model = "lobes"

# Load telescope
telescope = load_telescope(ms_path, element_response_model=response_model)
station_name = telescope.station_name(station_id)

if station_name not in ["CS302LBA", "SE607LBA"] and response_model == "lobes":
    logging.warning(
        f"LOBEs response requested, but not available for station {station_name}. EveryBeam defaults back to Hamaker"
    )

# Make coordinates for mesh
x_v = np.linspace(-1.0, 1.0, 128)
y_v = np.linspace(-1.0, 1.0, 128)

response = np.empty((y_v.size, x_v.size, 2, 2), dtype=np.cdouble)
response.fill(np.nan)
for i, x in enumerate(x_v):
    for j, y in enumerate(y_v):
        if (x**2 + y**2) <= 1.0:
            # Compute theta/phi and resulting direction vector
            theta = np.arcsin(np.sqrt(x * x + y * y))
            phi = np.arctan2(y, x)
            direction = thetaphi2cart(theta, phi)
            if mode == "element":
                response[j, i, :, :] = telescope.element_response(
                    time,
                    station_id,
                    element_id,
                    frequency,
                    direction,
                    is_local,
                    rotate=rotate,
                )
            elif mode == "station":
                response[j, i, :, :] = telescope.station_response(
                    time,
                    station_id,
                    frequency,
                    direction,
                    station0,
                    rotate=rotate,
                )
            else:
                raise Exception(
                    "Unrecognized response mode. Must be either station or element"
                )

# Plotting
X, Y = np.meshgrid(x_v, y_v)
label = ("X", "Y")
fig, axs = plt.subplots(2, 4, figsize=(10, 5), sharex=True, sharey=True)
fig.suptitle(f"{mode} response for station {station_name}: {response_model}")
for i, row in enumerate(range(2)):
    for j, col in enumerate(range(2)):
        im1 = axs[row, col].pcolor(X, Y, np.abs(response[:, :, row, col]))
        axs[row, col].set_aspect("equal", "box")
        axs[row, col].set_title(f"abs({label[j]}{label[i]})")
        fig.colorbar(im1, ax=axs[row, col])

        im2 = axs[row, col + 2].pcolor(
            X, Y, np.angle(response[:, :, row, col])
        )
        axs[row, col + 2].set_aspect("equal", "box")
        axs[row, col + 2].set_title(f"angle({label[j]}{label[i]})")
        fig.colorbar(im2, ax=axs[row, col + 2])

plt.show()



#!/usr/bin/env python3
# Copyright (C) 2020 ASTRON (Netherlands Institute for Radio Astronomy)
# SPDX-License-Identifier: GPL-3.0-or-later

import numpy as np
from scipy.constants import speed_of_light
from scipy.io import loadmat
from F4far_new import F4far_new
import lobes
import h5py


import matplotlib

matplotlib.use("tkagg")
from matplotlib import pyplot as plt

coords = loadmat("CS302_coords")

simdata = loadmat("LBA_CS302_fine")

# %
# % Compensate phase for element position
# %

P = coords["LBA"]["P"][0, 0][:, 0]
Q = coords["LBA"]["Q"][0, 0][:, 0]
positions = np.array([P, Q])

V_pol1 = simdata["Vdiff_pol1"]

theta = simdata["Theta"] / 180.0 * np.pi
phi = simdata["Phi"] / 180.0 * np.pi

Freq = simdata["Freq"][0, :]

K = 2.0 * np.pi * Freq / speed_of_light

sintheta = np.sin(theta)
cosphi = np.cos(phi)
sinphi = np.sin(phi)


x = sintheta.T * cosphi
y = sintheta.T * sinphi

dirs = np.array([x, y])

distance_diff = np.tensordot(positions, dirs, ((0,), (0,)))

phase_diff = np.multiply.outer(distance_diff, K)

phasor = np.exp(-1j * phase_diff)

V_pol1 *= phasor[:, np.newaxis, :, :, :]

El = 0
FreqN = 2

nr_elements = V_pol1.shape[0]
nr_frequencies = V_pol1.shape[-1]

Etheta = V_pol1[:, 0, 1:91, :-1, :].transpose((0, 3, 1, 2))
Ephi = V_pol1[:, 1, 1:91, :-1, :].transpose((0, 3, 1, 2))

Etheta = Etheta.reshape((nr_elements, nr_frequencies, -1))
Ephi = Ephi.reshape((nr_elements, nr_frequencies, -1))

E = np.concatenate((Etheta, Ephi), axis=2)

theta = theta[0, 1:91]
phi = phi[0, 0:-1]

theta_mesh, phi_mesh = np.meshgrid(theta, phi)

beta = 2.0 * np.pi * Freq[FreqN] / speed_of_light

Nmax = 21

theta_mesh_flat = theta_mesh.flatten()
phi_mesh_flat = phi_mesh.flatten()

F = np.zeros((E.shape[2], 2 * Nmax * (Nmax + 2)), dtype=np.complex128)

jn = 0
nms = []
for n in range(1, Nmax + 1):
    print(n)
    for m in range(-n, n + 1):
        for s in [1, 2]:
            # q2, q3 = F4far_new(s,m,n,theta_mesh_flat, phi_mesh_flat,beta);
            q2, q3 = lobes.F4far_new(
                s, m, n, theta_mesh_flat, phi_mesh_flat, beta
            )
            F[:, jn] = np.concatenate((q2, q3))
            nms.append((n, m, s))
            jn += 1

print("Inverting matrix")
Finv = np.linalg.pinv(F)

print("Compute product")
p = np.tensordot(Finv, E, axes=([1], [2]))

idx = np.argsort(np.mean(abs(p) ** 2, axis=(1, 2)))[::-1]

plt.figure()
plt.semilogy(abs(p[idx, :, :]).reshape((p.shape[0], -1)))

plt.show()

with h5py.File("LBA_CS302.hdf5", "w") as f:
    f.create_dataset("coefficients", data=p.T)
    f.create_dataset("nms", data=nms)
    f.create_dataset("frequencies", data=Freq.astype(np.double))

# disp(' ')
# Q=pinv(F)*E;
# Erecon=F*Q;
# E=reshape(E,length(Theta),length(Phi),2);
# Erecon=reshape(Erecon,length(Theta),length(Phi),2);
# %
# h1=figure;
# plot(Theta,20.*log10(abs(E(:,Phi==45,1))),'b')
# grid on
# hold on
# plot(Theta,20.*log10(abs(Erecon(:,Phi==45,1))),'r')
# xlabel('Theta (degrees)')
# ylabel('dB')
# title({['EEP element ' num2str(El) ', \phi=45 deg., freq. ' num2str(Freq(FreqN)./1e6) ' MHz'],'\theta-component (amplitude)'})
# h2=figure;
# plot(Theta,angle(E(:,Phi==45,1))./pi.*180,'b')
# grid on
# hold on
# plot(Theta,angle(Erecon(:,Phi==45,1))./pi.*180,'r')
# xlabel('Theta (degrees)')
# ylabel('degrees')
# title({['EEP element ' num2str(El) ', \phi=45 deg., freq. ' num2str(Freq(FreqN)./1e6) ' MHz'],'\theta-component (phase)'})
# h3=figure;
# plot(Theta,20.*log10(abs(E(:,Phi==135,2))),'b')
# grid on
# hold on
# plot(Theta,20.*log10(abs(Erecon(:,Phi==135,2))),'r')
# xlabel('Theta (degrees)')
# ylabel('dB')
# title({['EEP element 1, \phi=135 deg., freq. ' num2str(Freq(FreqN)./1e6) ' MHz'],'\phi-component (amplitude)'})
# h4=figure;
# plot(Theta,angle(E(:,Phi==135,2))./pi.*180,'b')
# grid on
# hold on
# plot(Theta,angle(Erecon(:,Phi==135,2))./pi.*180,'r')
# xlabel('Theta (degrees)')
# ylabel('degrees')
# title({['EEP element 1, \phi=135 deg., freq. ' num2str(Freq(FreqN)./1e6) ' MHz'],'\phi-component (phase)'})
# %
# Diff=Erecon./E;
# h5=figure;
# plot(Theta,20.*log10(abs(Diff(:,Phi==45,1))),'b')
# grid on
# hold on
# plot(Theta,20.*log10(abs(Diff(:,Phi==135,2))),'r')
# xlabel('Theta (degrees)')
# ylabel('dB')
# legend('\theta-comp., \phi=45^\circ','\phi-comp., \phi=135^\circ')
# title({'Difference between original and','reconstructed EEP (amplitude)', ...
# ['Element ' num2str(El) ', freq. ' num2str(Freq(FreqN)./1e6) ' MHz']})
# h6=figure;
# plot(Theta,angle(Diff(:,Phi==45,1))./pi.*180,'b')
# grid on
# hold on
# plot(Theta,angle(Diff(:,Phi==135,2))./pi.*180,'r')
# xlabel('Theta (degrees)')
# ylabel('degrees')
# legend('\theta-comp., \phi=45^\circ','\phi-comp., \phi=135^\circ')
# title({'Difference between original and','reconstructed EEP (phase)', ...
# ['Element ' num2str(El) ', freq. ' num2str(Freq(FreqN)./1e6) ' MHz']})
# %
# figure(h1)
# legend('original','reconstructed')
# figure(h2)
# legend('original','reconstructed')
# figure(h3)
# legend('original','reconstructed')
# figure(h4)
# legend('original','reconstructed')
# figure
# stem(abs(Q),'b')
# grid on
# xlabel('Index number')
# title('Absolute value of calculated coefficients')
# %
# disp(['Least square error = ' num2str(norm(E(:)-F*pinv(F)*E(:),2))])
# disp(['Condition number F = ' num2str(cond(F))])
# disp(['Size F = ' num2str(size(F))])
# disp(['Rank F = ' num2str(rank(F))])
# disp(' ')



# Copyright (C) 2021 ASTRON (Netherlands Institute for Radio Astronomy)
# SPDX-License-Identifier: GPL-3.0-or-later

from everybeam import load_telescope, OSKAR, GridSettings
import pytest
import os
import numpy as np

DATADIR = os.environ["DATA_DIR"]


@pytest.fixture
def oskar_setup():
    gs = GridSettings()
    gs.width = gs.height = 16
    gs.ra = 0.349066
    gs.dec = -0.523599
    gs.dl = gs.dm = 0.5 * np.pi / 180.0
    gs.l_shift = gs.m_shift = 0.0

    return {
        "filename": "OSKAR_MOCK.ms",
        "time": 4.45353e09,
        # Frequency of channel 4
        "freq": 5.0e07,
        "coordinate_system": gs,
        # Reference cpp solution for station 0, see cpp/test/toskar.cc
        "station_id": 0,
        "cpp_response": np.array(
            [
                [-0.00115441 + 0.00881435j, 6.07546e-05 + 0.0013592j],
                [-0.000379462 + 0.00158725j, 0.000982019 - 0.00856565j],
            ]
        ),
    }


@pytest.mark.parametrize("ref", [pytest.lazy_fixture("oskar_setup")])
def test_oskar(ref):
    ms_path = os.path.join(DATADIR, ref["filename"])
    differential_beam = False
    telescope = load_telescope(
        ms_path,
        use_differential_beam=differential_beam,
        element_response_model="skala40_wave",
    )

    assert type(telescope) is OSKAR
    response = telescope.station_response(
        ref["time"],
        ref["station_id"],
        ref["freq"],
        ref["coordinate_system"].ra,
        ref["coordinate_system"].dec,
    )

    # Reference solution same as pixel (8,8) in toskar.cc
    np.testing.assert_allclose(response, ref["cpp_response"], rtol=1e-5)



# Copyright (C) 2020 ASTRON (Netherlands Institute for Radio Astronomy)
# SPDX-License-Identifier: GPL-3.0-or-later

# Original Matlab code by Michel Arts
# Ported to python by Sebastiaan van der Tol
# Interface change: theta, phi are now in radians
# 2019-10-16

import numpy as np
from scipy.special import lpmn
import lobes


def F4far_new(s, m, n, theta, phi, beta):
    if s == 1:
        q1, q2, q3 = F41(m, n, theta, phi, beta)
        return q2, q3
    elif s == 2:
        q1, q2, q3 = F42(m, n, theta, phi, beta)
        return q2, q3
    raise RuntimeError(
        "Error arguments to F4far_new: s should be either 1 or 2"
    )


def legendre(n, x):
    results = np.zeros((n + 1, len(x)))
    for i in range(len(x)):
        assoc_legendre, assoc_legendre_derivative = lpmn(n, n, x[i])
        results[:, i] = assoc_legendre[:, -1]
    return results


def P(m, n, x):
    q = legendre(n, x)
    if n > 0:
        q = q[abs(m), :]

    q1 = lobes.P(m, n, x)

    if m < 0:
        q = (
            (-(1.0**-m))
            * np.math.factorial(n + m)
            / np.math.factorial(n - m)
            * q
        )

    return q


def Pacc(m, n, z):
    q = (
        -(n + m) * (n - m + 1) * np.sqrt(1 - z * z) * P(m - 1, n, z)
        - m * z * P(m, n, z)
    ) / (z * z - 1)
    q1 = lobes.Pacc(m, n, z)

    return q


def F41(m, n, theta, phi, beta):
    if m != 0:
        C = (
            beta
            * np.sqrt(60)
            * 1.0
            / np.sqrt(n * (n + 1))
            * (-m / abs(m)) ** m
        )
    else:
        C = beta * np.sqrt(60) * 1.0 / np.sqrt(n * (n + 1))
    q1 = np.zeros((1, len(theta)))
    q2 = (
        C
        * (-1j) ** (-n - 1)
        / beta
        * 1j
        * m
        / (np.sin(theta))
        * np.sqrt(
            (2.0 * n + 1)
            / 2
            * np.math.factorial(n - abs(m))
            / np.math.factorial(n + abs(m))
        )
        * P(abs(m), n, np.cos(theta))
        * np.exp(1j * m * phi)
    )
    q3 = (
        C
        * (-1j) ** (-n - 1)
        / beta
        * np.sqrt(
            (2.0 * n + 1)
            / 2.0
            * np.math.factorial(n - abs(m))
            / np.math.factorial(n + abs(m))
        )
        * Pacc(abs(m), n, np.cos(theta))
        * np.sin(theta)
        * np.exp(1j * m * phi)
    )
    return (q1, q2, q3)


def F42(m, n, theta, phi, beta):
    if m != 0:
        C = (
            beta
            * np.sqrt(60)
            * 1.0
            / np.sqrt(n * (n + 1))
            * (-m / abs(m)) ** m
        )
    else:
        C = beta * np.sqrt(60) * 1.0 / np.sqrt(n * (n + 1))

    q1 = np.zeros((1, len(theta)))
    q2 = (
        -C
        * (-1j) ** (-n)
        / beta
        * np.sqrt(
            (2.0 * n + 1)
            / 2.0
            * np.math.factorial(n - abs(m))
            / np.math.factorial(n + abs(m))
        )
        * Pacc(abs(m), n, np.cos(theta))
        * np.sin(theta)
        * np.exp(1j * m * phi)
    )
    q3 = (
        C
        * (-1j) ** (-n)
        / beta
        * 1j
        * m
        / np.sin(theta)
        * np.sqrt(
            (2.0 * n + 1)
            / 2.0
            * np.math.factorial(n - abs(m))
            / np.math.factorial(n + abs(m))
        )
        * P(abs(m), n, np.cos(theta))
        * np.exp(1j * m * phi)
    )

    return (q1, q2, q3)



import everybeam
import h5py
import numpy as np
import os

DATA_DIR = os.environ["DATA_DIR"]


def test_lwa_values():
    """Test that the E-fields calculated by Everybeam for LWA using spherical harmonics fitting is equal to the E-fields in the simulations file"""

    lwa = everybeam.ElementResponse.create("lwa")

    # Create a theta-phi grid containing some of the points in the
    # simulation file, so that the processing is faster
    phi_deg = [0, 18, 36, 54, 72, 90]
    theta_deg = [0, 40, 80, 120, 160, 200, 240, 280, 320, 360]
    phi = np.radians(phi_deg)
    theta = np.radians(theta_deg)
    phigrid, thetagrid = np.meshgrid(theta, phi)
    phi_plot = phigrid.flatten()
    theta_plot = thetagrid.flatten()

    # Pick only 3 frequencies to ensure a reasonable processing time
    freqs = np.array([1.0e07, 5.0e07, 1.0e08])

    pol_1_everybeam = np.zeros(
        (2, len(freqs), len(phi), len(theta)), dtype=complex
    )
    pol_2_everybeam = np.zeros(
        (2, len(freqs), len(phi), len(theta)), dtype=complex
    )

    station_id = np.full((len(phi) * len(theta), 1), 0)
    for freq_idx in np.arange(len(freqs)):
        freq_array = np.full((len(phi) * len(theta), 1), freqs[freq_idx])

        x = lwa.response(
            station_id.flatten(),
            freq_array.flatten(),
            theta_plot.flatten(),
            phi_plot.flatten(),
        )

        x_reshaped = x.reshape((len(phi), len(theta), 2, 2))

        pol_1_everybeam[0, freq_idx, :, :] = x_reshaped[:, :, 0, 0]
        pol_1_everybeam[1, freq_idx, :, :] = x_reshaped[:, :, 0, 1]
        pol_2_everybeam[0, freq_idx, :, :] = x_reshaped[:, :, 1, 0]
        pol_2_everybeam[1, freq_idx, :, :] = x_reshaped[:, :, 1, 1]

    data_path = os.path.join(DATA_DIR, "OVRO_LWA_FEKO_TEST.h5")
    h5 = h5py.File(data_path, "r")

    X_pol_e_theta = np.array(h5["X-pol_Efields/etheta"])
    X_pol_e_phi = np.array(h5["X-pol_Efields/ephi"])
    Y_pol_e_theta = np.array(h5["Y-pol_Efields/etheta"])
    Y_pol_e_phi = np.array(h5["Y-pol_Efields/ephi"])

    # Select the points in the simulation file corresponding to the phi-theta
    # grid used in everybeam
    X_pol_e_theta = X_pol_e_theta[:, phi_deg, :][:, :, theta_deg]
    X_pol_e_phi = X_pol_e_phi[:, phi_deg, :][:, :, theta_deg]
    Y_pol_e_theta = Y_pol_e_theta[:, phi_deg, :][:, :, theta_deg]
    Y_pol_e_phi = Y_pol_e_phi[:, phi_deg, :][:, :, theta_deg]

    pol_1_simulation = np.stack((X_pol_e_theta, X_pol_e_phi))
    pol_2_simulation = np.stack((Y_pol_e_theta, Y_pol_e_phi))

    np.testing.assert_allclose(
        pol_1_everybeam, pol_1_simulation, rtol=1e-10, atol=0.0015
    )
    np.testing.assert_allclose(
        pol_2_everybeam, pol_2_simulation, rtol=1e-10, atol=0.0016
    )



#!/usr/bin/env python3

"""
Script to un-publish and remove a Debian repository from a Debian package repository server.
"""

import argparse
import logging
import sys
from aptly_api import Client
from aptly_api.base import AptlyAPIException

logging.basicConfig(level=logging.INFO, format="%(levelname)s: %(message)s")
logger = logging.getLogger()


def parse_arguments():
    """
    Parse command-line arguments.
    :return: argparse.Namespace
    """
    # Do not add help automatically to prevent that two sections of optional arguments appear
    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter,
                                     description="Un-publish and remove a Debian repository "
                                                 "from a Debian package repository server.",
                                     add_help=False)
    required = parser.add_argument_group("required arguments")
    optional = parser.add_argument_group("optional arguments")
    required.add_argument("reponame",
                          help="name of the repository to remove")
    required.add_argument("-d", "--distribution", required=True,
                          help="distribution of the repository (e.g. bionic)")
    optional.add_argument("-h", "--help", action="help",
                          help="show this help message and exit")
    optional.add_argument("-p", "--prefix", default=None,
                          help="prefix of the repository (default: <reponame>")
    optional.add_argument("-u", "--url", default="http://packages.astron.nl:8080",
                          help="URL of repository server (default: %(default)s)")
    args = parser.parse_args()
    args.prefix = args.reponame if args.prefix is None else args.prefix
    return args


def main(args):
    """
    Main function. It creates a new repository on the repository server, based on the command line
    arguments that were supplied.
    :param args: Arguments as parsed by `parse_arguments()`.
    :return: Error status: 0: success; 1: error; 2: warning
    """
    # Open connection to the server
    aptly = Client(args.url)
    status = 0

    # Un-publish the repository
    try:
        aptly.publish.drop(prefix=args.prefix,
                        distribution=args.distribution)
        logger.info("Un-published repository %s", args.reponame)
    except AptlyAPIException as excp:
        logger.warning(excp)
        status = 2

    # Remove the repository from the server
    try:
        aptly.repos.delete(args.reponame)
        logger.info("Deleted repository %s", args.reponame)
    except AptlyAPIException as excp:
        logger.error(excp)
        status = 1

    return status


if __name__ == "__main__":
    sys.exit(main(parse_arguments()))



#!/usr/bin/env python3
# Copyright (C) 2020 ASTRON (Netherlands Institute for Radio Astronomy)
# SPDX-License-Identifier: GPL-3.0-or-later

""" Add beam info to OSKAR generated MeasurementSet """


import casacore.tables as pt
import pandas as pd
from lofarantpos import geo
from numpy.linalg import norm
import numpy as np
import logging
import sys


def read_telescope_center(oskar_telescope_dir: str):
    """Read telescope center from OSKAR metadata and convert to ITRF"""
    oskar_telescope_center = pd.read_csv(
        f"{oskar_telescope_dir}/position.txt", header=None, sep=" "
    )

    if len(oskar_telescope_center.columns) == 2:
        oskar_telescope_center["Height"] = [0.0]

    oskar_telescope_center.columns = ["Lon", "Lat", "Height"]
    telescope_center_itrf = geo.xyz_from_geographic(
        np.deg2rad(oskar_telescope_center["Lon"][0]),
        np.deg2rad(oskar_telescope_center["Lat"][0]),
        oskar_telescope_center["Height"][0],
    )
    return telescope_center_itrf


def fix_antenna(oskar_ms_name: str, telescope_center_itrf: np.array):
    """Make POSITION in ::ANTENNA subtable absolute"""
    anttable = pt.table(f"{oskar_ms_name}::ANTENNA", readonly=False, ack=False)
    if np.isclose(norm(anttable[0]["POSITION"]), 0, atol=1e-3):
        anttable.putcol(
            "POSITION", anttable.getcol("POSITION") + telescope_center_itrf
        )
    else:
        logging.info("Positions are already absolute")
    anttable.close()


def add_array_center(oskar_ms_name: str, telescope_center_itrf: np.array):
    """Add ARRAY_CENTER column to ::OBSERVATION subtable"""
    obstable = pt.table(
        f"{oskar_ms_name}::OBSERVATION", readonly=False, ack=False
    )
    if "ARRAY_CENTER" not in obstable.colnames():
        anttable = pt.table(f"{oskar_ms_name}::ANTENNA", ack=False)
        coldesc = anttable.getcoldesc("POSITION")
        coldesc["name"] = "ARRAY_CENTER"
        coldesc["comment"] = "Reference position for array"
        obstable.addcols(coldesc)

    obstable.putcol("ARRAY_CENTER", np.array([telescope_center_itrf]))
    obstable.close()


def add_phased_array_table(oskar_ms_name: str):
    """Add PHASED_ARRAY subtable to measurement set"""

    anttable = pt.table(f"{oskar_ms_name}::ANTENNA", ack=False)
    phasedarraytable = pt.table(
        f"{oskar_ms_name}/PHASED_ARRAY",
        pt.maketabdesc([]),
        nrow=anttable.nrows(),
    )

    oskar_ms = pt.table(f"{oskar_ms_name}", readonly=False, ack=False)
    oskar_ms.putkeyword("PHASED_ARRAY", phasedarraytable, makesubrecord=True)
    oskar_ms.close()

    position_coldesc = anttable.getcoldesc("POSITION")
    position_coldesc["comment"] = "Position of antenna ﬁeld"
    position_coldesc["name"] = "POSITION"
    phasedarraytable.addcols(position_coldesc)

    coordinate_system_coldesc = pt.makearrcoldesc(
        "COORDINATE_AXES",
        0.0,
        shape=[3, 3],
        comment="Local coordinate system",
        valuetype="double",
        keywords={
            "QuantumUnits": ["m", "m", "m"],
            "MEASINFO": {"Ref": "ITRF", "type": "direction"},
        },
    )
    phasedarraytable.addcols(coordinate_system_coldesc)
    pt.taql("UPDATE $phasedarraytable SET COORDINATE_AXES=0.")

    element_offset_coldesc = pt.makearrcoldesc(
        "ELEMENT_OFFSET",
        0.0,
        ndim=2,
        comment="Offset per element",
        valuetype="double",
        keywords={
            "QuantumUnits": ["m", "m", "m"],
            "MEASINFO": {"type": "position", "Ref": "ITRF"},
        },
    )
    phasedarraytable.addcols(element_offset_coldesc)

    element_flag_coldesc = pt.makearrcoldesc(
        "ELEMENT_FLAG",
        0.0,
        ndim=2,
        comment="Offset per element",
        valuetype="bool",
    )
    phasedarraytable.addcols(element_flag_coldesc)
    phasedarraytable.close()


def fill_phased_array(oskar_ms_name: str, oskar_telescope_dir: str):
    """Fill the ::PHASED_ARRAY subtable with info from the OSKAR directory"""
    element_locations = pd.read_csv(
        f"{oskar_telescope_dir}/station/layout.txt", header=None, sep=" "
    )
    if len(element_locations.columns) == 2:
        element_locations["Z"] = np.zeros_like(element_locations[0])

    oskar_telescope_center = pd.read_csv(
        f"{oskar_telescope_dir}/position.txt", header=None, sep=" "
    )
    oskar_telescope_center.columns = ["Lon", "Lat"]
    telescope_center_itrf = read_telescope_center(oskar_telescope_dir)

    normal_vector_ellipsoid = geo.normal_vector_ellipsoid(
        np.deg2rad(oskar_telescope_center["Lon"][0]),
        np.deg2rad(oskar_telescope_center["Lat"][0]),
    )

    local_to_itrf_projection_matrix = geo.projection_matrix(
        telescope_center_itrf, normal_vector_ellipsoid
    )

    element_locations_itrf = (
        local_to_itrf_projection_matrix @ element_locations.to_numpy().T
    )
    nr_pol = 2
    all_unflagged = np.zeros((nr_pol, element_locations_itrf.shape[1]))

    phasedarraytable = pt.table(
        f"{oskar_ms_name}::PHASED_ARRAY", readonly=False, ack=False
    )

    for stationnr in range(len(phasedarraytable)):
        phasedarraytable.putcell(
            "COORDINATE_AXES", stationnr, local_to_itrf_projection_matrix.T
        )
        phasedarraytable.putcell(
            "ELEMENT_OFFSET", stationnr, element_locations_itrf
        )
        phasedarraytable.putcell("ELEMENT_FLAG", stationnr, all_unflagged)

    anttable = pt.table(f"{oskar_ms_name}::ANTENNA", ack=False)
    phasedarraytable.putcol("POSITION", anttable.getcol("POSITION"))
    phasedarraytable.close()


def main(oskar_ms_name: str, oskar_telescope_dir: str):
    """Add beam info to OSKAR generated MeasurementSet"""
    telescope_center_itrf = read_telescope_center(oskar_telescope_dir)
    fix_antenna(oskar_ms_name, telescope_center_itrf)
    add_array_center(oskar_ms_name, telescope_center_itrf)
    add_phased_array_table(oskar_ms_name)
    fill_phased_array(oskar_ms_name, oskar_telescope_dir)


if __name__ == "__main__":
    main(sys.argv[1], sys.argv[2])



#!/usr/bin/env python3
# Copyright (C) 2021 ASTRON (Netherlands Institute for Radio Astronomy)
# SPDX-License-Identifier: GPL-3.0-or-later

"""
    Script converts LOBEs simulation data to a coefficient file, by fitting spherical harmonics to
    the simulated data. In order to run this script, please make sure
    that the everybeam shared library is on your LD_LIBRARY_PATH and on your PYTHONPATH, See below for
    example export commands:
    - export LD_LIBRARY_PATH=~/opt/everybeam/lib:$LD_LIBRARY_PATH
    - export PYTHONPATH=/home~/opt/everybeam/lib/python3.6/site-packages:$PYTHONPATH

    Script can handle two ways of providing the input data:

    1. station coordinates and simulated results for all frequencies are provided in one HDF5 file
    2. simulated results are provided in a separate MATLAB file per frequency. Station coordinates
    are provided in a separate MATLAB file or retrieved using the lofarantpos tool.
"""

import numpy as np
from scipy.constants import speed_of_light
from scipy.io import loadmat
from everybeam.lobes import F4far_new
import h5py
import os, glob


def read_lofar_antenna_positions(station):
    """
    Get antenna positions in local p,q coordinate system [m]
    within LOFAR station using lofarantpos tool.

    Please note that this function requires lofarantpos to be installed.
    https://pypi.org/project/lofarantpos/

    Parameters
    ----------
    station : str
        LOFAR station name

    Returns
    -------
    np.2darray
        Numpy array (2, nantenna) of p,q-coordinates [m]
    """

    from lofarantpos.db import LofarAntennaDatabase

    db = LofarAntennaDatabase()
    return db.antenna_pqr(station)[:, :2].T


def read_coordinates_from_file(path):
    """
    Read p,q-coordinates from coordinate file at
    specified input path.

    Parameters
    ----------
    path : str
        Path to directory containing simulated results

    Returns
    -------
    np.2darray
        2D array with coordinates in local east-north system [m]
    """

    coord_file = glob.glob(f"{path}/*_coords.mat")
    assert len(coord_file) == 1
    coords = loadmat(coord_file[0])
    p = coords["LBA"]["P"][0, 0].flatten()
    q = coords["LBA"]["Q"][0, 0].flatten()
    return np.array([p, q])


def read_simulation_files(path):
    """
    Read simulation files, where it is assumed that
    each simulation file contains the results for one
    frequency.

    Parameters
    ----------
    path : str
        Path to directory containing simulated results

    Returns
    -------
    dict
        Dictionary, containing the following fields:
        - theta: zenith angles of simulated results [rad]
        - phi: elevation angle of simulated results [rad]
        - freq, np.1darray of frequencies [MHz]
        - v_pol1, returned shape is (#antenna elements, 2, #theta, #phi, #freqs)
        - v_pol2, returned shape is (#antenna elements, 2, #theta, #phi, #freqs)
    """

    file_list = []
    for file_path in glob.glob(f"{path}/*mhz.mat"):
        file_list.append(file_path)

    nfreqs = len(file_list)
    freqs = np.zeros(nfreqs)

    # Hard coded for now, could be inferred when reading the files
    # Alternatively, use np.stack?!
    v_pol1 = np.empty((96, 2, 91, 73, nfreqs), dtype=np.complex128)
    v_pol2 = np.empty((96, 2, 91, 73, nfreqs), dtype=np.complex128)

    for i, file_path in enumerate(file_list):
        print(f"Reading file path {file_path}: {i+1}/{nfreqs}")
        simdata = loadmat(file_path)
        freqs[i] = simdata["Freq"][0, 0]

        if i == 0:
            # Convert to radians
            theta = simdata["Theta"] / 180.0 * np.pi
            phi = simdata["Phi"] / 180.0 * np.pi
        else:
            np.testing.assert_allclose(theta, simdata["Theta"] / 180.0 * np.pi)
            np.testing.assert_allclose(phi, simdata["Phi"] / 180.0 * np.pi)

        v_pol1[..., i] = simdata["Vdiff_pol1"]
        v_pol2[..., i] = simdata["Vdiff_pol2"]

    if np.any(freqs[:-1] > freqs[1:]):
        # Then we have to reorder frequencies - and simulated data -
        # ascendingly
        idcs = np.argsort(freqs)
        freqs = freqs[idcs]
        v_pol1 = v_pol1[..., idcs]
        v_pol2 = v_pol2[..., idcs]

    theta = theta.flatten()
    phi = phi.flatten()

    # Return as dictionary
    # Note: the phi axis runs from [0, 2pi> (not including 2 pi) by cutting off
    # the last element. This avoids 0 and 2pi to be weighted twice.
    return {
        "theta": theta,
        "phi": phi[:-1],
        "v_pol1": v_pol1[..., :-1, :],
        "v_pol2": v_pol2[..., :-1, :],
        "freq": freqs,
    }


def read_h5_file(sim_path, sim_file, freq_select=None, element_select=None):
    """
    Read h5 file with LOBEs simulated data. Assumes the following fields to be present:
        - "X" : east coordinate of antenna element [m]
        - "Y" : north coordinate of antenna element [m]
        - "Freq": frequencies [MHz]
        - "Theta" : zenith angles for which the embedded element patterns are calculated [deg]
        - "Phi" : elevation angles for which the embeeded element patterns are calculated [deg]
        - "Vdiff_pol1" : differential voltage for polarization 1, assumed shape (#freqs, #phi, #theta, 2, #antenna elements)
        - "Vdiff_pol2" : differential voltage for polarization 2, assumed shape (#freqs, #phi, #theta, 2, #antenna elements)

    Parameters
    ----------
    sim_path : str
        Path containing the simulated data
    sim_file : str
        File name containing the simulated data

    Returns
    -------
    dict
        Dictionary containing numpy arrays of the simulated data. Contains the following keys:
            - coords
            - theta, zenith angle [rad]
            - phi, elevation angle [rad]
            - freq, np.1darray of frequencies [MHz]
            - v_pol1, returned shape is (#antenna elements, 2, #theta, #phi, #freqs)
            - v_pol2, returned shape is (#antenna elements, 2, #theta, #phi, #freqs)
    """

    with h5py.File(os.path.join(sim_path, sim_file), "r") as f:
        X = f["X"] if element_select is None else f["X"][:, element_select]
        Y = f["Y"] if element_select is None else f["Y"][:, element_select]
        coords = np.concatenate((X, Y), axis=0)

        theta = np.deg2rad(np.array(f["Theta"]).flatten())
        phi = np.deg2rad(np.array(f["Phi"]).flatten())

        freq = (
            np.array(f["Freq"]).flatten()
            if freq_select is None
            else np.array(f["Freq"]).flatten()[freq_select]
        )

        if freq_select is None and element_select is None:
            v_pol1_ref = f["Vdiff_pol1"]
            v_pol2_ref = f["Vdiff_pol2"]
        else:
            v_pol1_ref = f["Vdiff_pol1"][freq_select, ..., element_select]
            v_pol2_ref = f["Vdiff_pol2"][freq_select, ..., element_select]

        # Compose complex valued array and transpose
        v_pol1 = (v_pol1_ref["real"] + 1j * v_pol1_ref["imag"]).transpose(
            (4, 3, 2, 1, 0)
        )
        v_pol2 = (v_pol2_ref["real"] + 1j * v_pol2_ref["imag"]).transpose(
            (4, 3, 2, 1, 0)
        )

    # Note: the phi axis runs from [0, 2pi> (not including 2 pi) by cutting off
    # the last element. This avoids 0 and 2pi to be weighted twice.
    return {
        "coords": coords,
        "theta": theta,
        "phi": phi[:-1],
        "freq": freq,
        "v_pol1": v_pol1[..., :-1, :],
        "v_pol2": v_pol2[..., :-1, :],
    }


def extract_electromagnetic_field(V_pol, weights=None):
    """
    Extract the simulated electromagnetic field.

    Parameters
    ----------
    V_pol : np.ndarray
        Differential voltage, shape expected to be
        (#elements, 2, #theta, #phi, #freqs)
    weights : np.1darray
        1darray with (zenith angle) weights with size equal to #theta, by default None

    Returns
    -------
    np.ndarray, np.ndarray/None
        Unweighted electromagnetic field, Weighted electromagnetic field (None if weights is None).
        Shape equals (#element, #freqs, #phi * #theta)
    """
    nr_elements = V_pol.shape[0]
    nr_frequencies = V_pol.shape[-1]

    Etheta = V_pol[:, 0, ...].transpose((0, 3, 2, 1))
    Ephi = V_pol[:, 1, ...].transpose((0, 3, 2, 1))

    if weights is not None:
        Etheta_w = Etheta * weights
        Ephi_w = Ephi * weights

        Etheta_w = Etheta_w.reshape((nr_elements, nr_frequencies, -1))
        Ephi_w = Ephi_w.reshape((nr_elements, nr_frequencies, -1))

    Etheta = Etheta.reshape((nr_elements, nr_frequencies, -1))
    Ephi = Ephi.reshape((nr_elements, nr_frequencies, -1))

    if weights is not None:
        return (
            np.concatenate((Etheta, Ephi), axis=2),
            np.concatenate((Etheta_w, Ephi_w), axis=2),
        )
    else:
        return np.concatenate((Etheta, Ephi), axis=2), None


def apply_phasor(v_pol1, v_pol2, freqs, positions, x, y):
    """
    Apply phasor to simulated electromagnetic field.

    Parameters
    ----------
    v_pol1 : np.ndarray
        Differential voltage X-polarization,
        (#elements, 2, #theta, #phi, #freqs)
    v_pol2 : np.ndarray
        Differential voltage X-polarization,
        (#elements, 2, #theta, #phi, #freqs)
    freqs : np.1darray
        Frequencies [Hz]
    positions : np.2darray
        Local (p,q)-coordinates of elements [m]
    x : np.1darray
        Cartesian x-coordinate of (theta, phi) system
    y : np.1darray
        Cartesion y-coordinate of (theta, phi) system
    """

    K = 2.0 * np.pi * freqs / speed_of_light
    dirs = np.array([x, y])

    distance_diff = np.tensordot(positions, dirs, ((0,), (0,)))
    phase_diff = np.multiply.outer(distance_diff, K)
    phasor = np.exp(-1j * phase_diff)

    v_pol1[:] *= phasor[:, np.newaxis, ...]
    v_pol2[:] *= phasor[:, np.newaxis, ...]


def compute_polyomial_expansion(theta, phi, nmax, weights=None):
    """
    Compute the (Legendre) polynomial expansion

    Parameters
    ----------
    theta : np.1darray
        1d array of zenith angles [rad]
    phi : np.1darray
        1d array of elevation angles [rad]
    nmax : int
        Number/order of spherical harmonics
    weights : np.1darray, optional
        1d of zenith angle weights, should have same size as theta array.
        Defaults to None, in which case no weight

    Returns
    -------
    np.2darray, np.2darray, np.ndarray
        Basis matrix, Weighted basis matrix, nms matrix. If weights is None,
        the weighted basis matrix will be None
    """

    Theta, Phi = np.meshgrid(theta, phi)
    F = np.zeros((2 * Theta.size, 2 * nmax * (nmax + 2)), dtype=np.complex128)

    jn = 0
    nms = []
    for n in range(1, nmax + 1):
        for m in range(-n, n + 1):
            for s in [1, 2]:
                q2, q3 = F4far_new(s, m, n, Theta.flatten(), Phi.flatten())
                F[:, jn] = np.concatenate((q2, q3))
                nms.append((n, m, s))
                jn += 1

    if weights is not None:
        weights = np.tile(weights, 2 * phi.size)
        F_w = F * weights.reshape(-1, 1)
        return F, F_w, nms
    else:
        return F, None, nms


def write_h5file(outfile, coeffs, nms, freqs):
    """
    Write results to h5 file

    Parameters
    ----------
    outfile : str
        Output file name (or path)
    coeffs : np.ndarray
        5d numpy array, expected dimensions (2, #freqs, #elements, #coefficients)
    nms : np.ndarray
    freqs : np.1darray
        1D array with frequencies
    outdir : str, optional
        Output directory, by default "."
    """
    with h5py.File(outfile, "w") as f:
        f.create_dataset("coefficients", data=coeffs)
        f.create_dataset("nms", data=nms)
        f.create_dataset("frequencies", data=freqs.astype(np.float64))


def make_plot(E, theta, phi, vmin=None, vmax=None):
    """
    Make plot of field E on x, y plane

    Parameters
    ----------
    E : np.ndarray
        Electromagnetic field
    theta : np.1darray
        1d array of zenith angle coordinates [rad]
    phi : np.1darray
        1d array of elevation angle coordinates [rad]
    vmin : float, optional
        Minimum value of color range
    vmax : float, optional
        Maximum value of color range
    """
    import matplotlib

    matplotlib.use("tkagg")
    from matplotlib import pyplot as plt

    theta_mesh, phi_mesh = np.meshgrid(theta, phi)
    r = np.sin(theta_mesh)
    xm = r * np.cos(phi_mesh)
    ym = r * np.sin(phi_mesh)

    plt.pcolor(xm, ym, E, vmin=vmin, vmax=vmax)
    plt.colorbar()
    plt.show()


def fit_and_write_lobes_coeffs(
    theta,
    phi,
    freq,
    positions,
    v_pol1,
    v_pol2,
    nmax,
    outfile,
    apply_weighting=False,
):
    """
    Fit and write coefficients for simulated lobes data

    Parameters
    ----------
    theta : np.1darray
        1d array of zenith angle coordinates [rad]
    phi : np.1darray
        1d array of elevation angle coordinates [rad]
    freq : np.1darray
        Array with frequencies [Hz]
    positions : np.2darray
        Element coordinates in local east-north system [m]
    v_pol1 : np.5darray
        Simulated voltage pattern for X-polarisation, is expected to have
        shape (#antenna elements, 2, #theta, #phi, #freqs)
    v_pol2 : np.5darray
        Simulated voltage pattern for X-polarisation, is expected to have
        shape (#antenna elements, 2, #theta, #phi, #freqs)
    nmax : int
        Number/order of spherical harmonics
    outfile : str
        Path to output h5 file
    apply_weighting : bool, optional
        Apply weighting with the sine of the zenith angle, i.e. sin(theta), by default False

    Returns
    -------
    np.2darray, np.ndarray
        Return basis function expansion and coefficients to ease further postprocessing
    """

    # Frequencies are in MHz, convert to Hz
    sintheta = np.sin(theta.reshape(-1, theta.size))
    weights = None if apply_weighting is None else sintheta.flatten()

    cosphi = np.cos(phi.reshape(-1, phi.size))
    sinphi = np.sin(phi.reshape(-1, phi.size))

    x = sintheta.T * cosphi
    y = sintheta.T * sinphi

    # NOTE: Always check that freqs are in Hz
    if positions is not None:
        apply_phasor(v_pol1, v_pol2, freq, positions, x, y)

    # X/Y-polarization
    E1, E1w = extract_electromagnetic_field(v_pol1, weights=weights)
    E2, E2w = extract_electromagnetic_field(v_pol2, weights=weights)

    F, F_w, nms = compute_polyomial_expansion(
        theta, phi, nmax, weights=weights
    )
    print("Inverting matrix")
    Finv = np.linalg.pinv(F_w) if F_w is not None else np.linalg.pinv(F)

    print("Compute coefficints as Finv \dot E")
    p1 = np.tensordot(Finv, E1w if E1w is not None else E1, axes=([1], [2]))
    p2 = np.tensordot(Finv, E2w if E2w is not None else E2, axes=([1], [2]))

    # Stack into 4D array
    p = np.stack((p1, p2), axis=-1)
    write_h5file(outfile, p.T, nms, freq)
    return F, p.T


def main():
    """
    There are two routes for fitting coefficients to the
    LOBEs simulations:
    - Route 1: station coordinates and simulated results for all frequencies are provided in one HDF5 file
    - Route 2: simulated results are provided in a separate HDF5 file per frequency. Also, station coordinates
    are provided in a separate HDF5 file.

    Examples for both routes - with paths and filenames on my local system - are provided below.
    """

    # Order of Legendre polynomial to be fitted. Adjust to suit your needs
    nmax = 21
    approach = "separate_hdf5_file"

    def single_hdf5_file():
        """
        Coordinates and simulated results for all frequencies in one HDF5 file
        """

        # Adjust to local paths
        station = "SE607"
        simpath = "/var/scratch/maljaars/data"
        simfile = f"{station}_patterns.mat"
        outfile = f"LOBES_{station}LBA.h5"
        outfile = os.path.join(".", outfile)
        freq_select = None
        element_select = None

        h5data = read_h5_file(
            simpath,
            simfile,
            freq_select=freq_select,
            element_select=element_select,
        )
        # Please note that freqs have to be converted from MHz to Hz!
        (
            F,
            pT,
        ) = fit_and_write_lobes_coeffs(
            h5data["theta"],
            h5data["phi"],
            h5data["freq"] * 1e6,
            h5data["coords"],
            h5data["v_pol1"],
            h5data["v_pol2"],
            nmax,
            outfile,
            apply_weighting=False,
        )
        return simdata["theta"], simdata["phi"], F, pT

    def separate_hdf5_files():
        """
        Simulated results for each frequency stored in separate file. Station coordinates
        stored in separate HDF5 file.
        """

        # Adjust to correct local paths
        station = "CS007"
        path = f"/var/scratch/maljaars/data/LOBES/ftp.astron.nl/outgoing/arts/{station}"
        outpath = "/var/scratch/maljaars/data/LOBES/core-station-coeffs"
        outfile = f"LOBES_{station}LBA.h5"
        outfile = os.path.join(outpath, outfile)

        # Read positions from file
        # positions = read_coordinates_from_file(path)
        # Use lofarantpos to read coordinates
        positions = read_lofar_antenna_positions(f"{station}LBA")
        simdata = read_simulation_files(path)

        (
            F,
            pT,
        ) = fit_and_write_lobes_coeffs(
            simdata["theta"],
            simdata["phi"],
            simdata["freq"],
            positions,
            simdata["v_pol1"],
            simdata["v_pol2"],
            nmax,
            outfile,
            apply_weighting=False,
        )
        return simdata["theta"], simdata["phi"], F, pT

    theta, phi, F, pT = (
        single_hdf5_file()
        if approach == "single_hdf5_file"
        else separate_hdf5_files()
    )

    # Plot modulus of fitted response for XX, freq 0, and element 0
    make_plot(
        np.abs(
            np.dot(F[0 : theta.size * phi.size, :], pT[0, 0, 0, :]).reshape(
                phi.size, theta.size
            )
        ),
        theta,
        phi,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Copyright (C) 2020 ASTRON (Netherlands Institute for Radio Astronomy)
# SPDX-License-Identifier: GPL-3.0-or-later

# Script to convert an ASCII beam model coefficient file to a .cc file for
# inclusion in the library. Whenever the beam model coefficients file are
# updated, this script can be run to automatically update the corresponding .cc
# files.
#
# $Id$

import sys
import time
import re


def flat_index(shape, index):
    """
    Compute the flat index of the element with the provided (N-dimensional)
    index of an N-dimensional array with the provided shape. Row-major order
    (or "C"-order) is assumed in the computation of the flattened index. The
    index is range checked against the provided shape.
    """

    assert len(shape) == len(index)

    if len(shape) == 0:
        return 0

    assert index[0] < shape[0]
    flat = index[0]

    for i in range(1, len(shape)):
        assert index[i] < shape[i]
        flat *= shape[i]
        flat += index[i]

    return flat


def regex(name, type, signed=True):
    """
    Return a regular expression to match a (possibly signed) int or float, using
    the named group syntax. The matching group will by assigned the provided
    name.
    """

    expr = None
    if type == "int":
        expr = "\d+"
    elif type == "float":
        expr = "(\d+(\.\d*)?|\.\d+)([eE][-+]?\d+)?"

    assert expr, "unknown type: %s" % type

    if signed:
        expr = "[-+]?" + expr

    return "(?P<%s>%s)" % (name, expr)


def main(args):
    """
    Main function for the conversion of of a coefficient file to a *.cc file. Typical usage is
    ./convert_coeff.py [FILE_IN] [FILE_OUT] [TYPE]

    Examples:

    ./convert_coeff.py element_beam_LBA.coeff defaultcoefflba.cc default_lba
    ./convert_coeff.py element_beam_HBA.coeff defaultcoeffhba.cc default_hba
    """
    print(
        "converting %s -> %s (variable name: %s)" % (args[0], args[1], args[2])
    )

    HEADER, COEFF = range(2)
    state = HEADER

    shape = None
    coeff = None
    freqAvg = None
    freqRange = None

    line_no = 0
    count = 0

    fin = file(args[0], "r")
    for line in fin:
        # Remove leading and trailing whitespace.
        line = line.strip()

        # Skip empty lines and comments.
        if len(line) == 0 or line[0] == "#":
            line_no += 1
            continue

        # Parse header information.
        if state == HEADER:
            match = re.match(
                "^d\s+%s\s+k\s+%s\s+pwrT\s+%s\s+pwrF\s+%s\s+"
                "freqAvg\s+%s\s+freqRange\s+%s$"
                % (
                    regex("d", "int", False),
                    regex("k", "int", False),
                    regex("pwrT", "int", False),
                    regex("pwrF", "int", False),
                    regex("freqAvg", "float", False),
                    regex("freqRange", "float", False),
                ),
                line,
            )
            assert match, 'unable to parse header: "%s"' % line

            shape = (
                int(match.group("k")),
                int(match.group("pwrT")),
                int(match.group("pwrF")),
                int(match.group("d")),
            )
            assert shape[3] == 2, "unsupported array shape, expected d == 2"

            size = reduce(lambda x, y: x * y, shape)
            print("coefficient array shape:", shape, "(%d total)" % size)

            freqAvg = match.group("freqAvg")
            freqRange = match.group("freqRange")

            coeff = [None for x in range(size)]
            state = COEFF

        # Parse coefficients.
        elif state == COEFF:
            match = re.match(
                "^%s\s+%s\s+%s\s+%s\s+%s\s+%s$"
                % (
                    regex("d", "int", False),
                    regex("k", "int", False),
                    regex("pwrT", "int", False),
                    regex("pwrF", "int", False),
                    regex("re", "float"),
                    regex("im", "float"),
                ),
                line,
            )
            assert match, "unable to parse line #%d" % line_no

            d = int(match.group("d"))
            k = int(match.group("k"))
            pwrT = int(match.group("pwrT"))
            pwrF = int(match.group("pwrF"))

            index = flat_index(shape, (k, pwrT, pwrF, d))
            coeff[index] = "std::complex<double>(%s, %s)" % (
                match.group("re"),
                match.group("im"),
            )

            count += 1

        line_no += 1

    fin.close()
    assert not (coeff is None) and all(coeff)

    # Write the output.
    fout = file(args[1], "w")

    print >> fout, "// Beam model coefficients converted by convert_coeff.py."
    print >> fout, "// Conversion performed on %s UTC using: " % time.strftime(
        "%Y/%m/%d/%H:%M:%S", time.gmtime()
    )
    print >> fout, "//     convert_coeff.py %s %s %s" % (
        args[0],
        args[1],
        args[2],
    )
    print >> fout
    print >> fout, "#include <complex>"
    print >> fout
    print >> fout, "// Center frequency, and frequency range for which the beam model coefficients"
    print >> fout, "// are valid. The beam model is parameterized in terms of a normalized"
    print >> fout, "// frequency f' in the range [-1.0, 1.0]. The appropriate conversion is:"
    print >> fout, "//"
    print >> fout, "//     f' = (f - center) / range"
    print >> fout, "//"
    print >> fout, "const double %s_freq_center = %s;" % (args[2], freqAvg)
    print >> fout, "const double %s_freq_range = %s;" % (args[2], freqRange)
    print >> fout
    print >> fout, "// Shape of the coefficient array: %dx%dx%dx2 (the size of the last dimension is" % (
        shape[0],
        shape[1],
        shape[2],
    )
    print >> fout, "// implied, and always equal to 2)."
    print >> fout, "//"
    print >> fout, "const unsigned int %s_coeff_shape[3] = {%d, %d, %d};" % (
        args[2],
        shape[0],
        shape[1],
        shape[2],
    )
    print >> fout
    print >> fout, '// The array of coefficients in row-major order ("C"-order).'
    print >> fout, "//"
    print >> fout, "const std::complex<double> %s_coeff[%d] = {" % (
        args[2],
        len(coeff),
    )

    i = 0
    while i < len(coeff):
        if i + 2 < len(coeff):
            print >> fout, "    %s, %s," % (coeff[i], coeff[i + 1])
            i += 2
        elif i + 2 == len(coeff):
            print >> fout, "    %s, %s" % (coeff[i], coeff[i + 1])
            i += 2
        else:
            print >> fout, "    %s" % coeff[i]
            i += 1

    print >> fout, "};"

    fout.close()


if __name__ == "__main__":
    if len(sys.argv) != 4:
        print(
            "convert a beam model coefficient (.coeff) file to a C++ (.cc) file."
        )
        print(
            "usage: convert_coeff.py <input-file> <output-file> <variable-name>"
        )
        sys.exit(1)

    main(sys.argv[1:])



#!/usr/bin/env python3
# Copyright (C) 2021 ASTRON (Netherlands Institute for Radio Astronomy)
# SPDX-License-Identifier: GPL-3.0-or-later

"""
    Script converts OVRO-LWA simulation data to a coefficient file, by fitting spherical harmonics to
    the simulated data. In order to run this script, please make sure
    that the everybeam shared library is on your LD_LIBRARY_PATH and on your PYTHONPATH, See below for
    example export commands:
    - export LD_LIBRARY_PATH=~/opt/everybeam/lib:$LD_LIBRARY_PATH
    - export PYTHONPATH=/home~/opt/everybeam/lib/python3.6/site-packages:$PYTHONPATH

    You can run this script as follows:
    python3 scripts/coeff_scripts/convert_lwa.py  --simulation_path LWA_SIMULATION.h5 --destination_path LWA_COEFFICIENTS.h5
"""
import argparse
import numpy as np
import h5py
from convert_lobes import fit_and_write_lobes_coeffs
import os


def read_lwa_simulation_file(path):
    """
    Read simulation files, where all frequencies are contained in a single HDF5 file

    Parameters
    ----------
    path : str
        Path to directory containing simulated results

    Returns
    -------
    dict
        Dictionary, containing the following fields:
        - theta: zenith angles of simulated results [rad]
        - phi: elevation angle of simulated results [rad]
        - freq, np.1darray of frequencies [MHz]
        - v_pol1, returned shape is (#antenna elements, 2, #theta, #phi, #freqs)
        - v_pol2, returned shape is (#antenna elements, 2, #theta, #phi, #freqs)
    """

    h5 = h5py.File(path, "r")

    freqs = np.array(h5["Freq(Hz)"])

    # Read polarizations from FEKO simulation file
    # For compatibility with the LOBES coefficients, add one extra dimension for the stations.
    # Since the data is the same for all stations, the dimension of this new axis is 1.
    v_pol1 = np.expand_dims(
        np.stack(
            (
                np.array(h5["X-pol_Efields/etheta"]),
                np.array(h5["X-pol_Efields/ephi"]),
            )
        ),
        axis=0,
    )
    v_pol2 = np.expand_dims(
        np.stack(
            (
                np.array(h5["Y-pol_Efields/etheta"]),
                np.array(h5["Y-pol_Efields/ephi"]),
            )
        ),
        axis=0,
    )

    # reshape from shape (#antenna elements, 2, #freqs, #theta, #phi)
    #           to shape (#antenna elements, 2, #theta, #phi, #freqs)
    v_pol1 = np.transpose(v_pol1, (0, 1, 3, 4, 2))
    v_pol2 = np.transpose(v_pol2, (0, 1, 3, 4, 2))

    # Theta and phi are already in radians
    theta = np.array(h5["theta_pts"])
    phi = np.array(h5["phi_pts"])

    if np.any(freqs[:-1] > freqs[1:]):
        # Then we have to reorder frequencies - and simulated data -
        # ascendingly
        idcs = np.argsort(freqs)
        freqs = freqs[idcs]
        v_pol1 = v_pol1[..., idcs]
        v_pol2 = v_pol2[..., idcs]

    theta = theta.flatten()
    phi = phi.flatten()

    # Return as dictionary
    # Note: the phi axis runs from [0, 2pi> (not including 2 pi) by cutting off
    # the last element. This avoids 0 and 2pi to be weighted twice.
    return {
        "theta": theta,
        "phi": phi[:-1],
        "v_pol1": v_pol1[..., :-1, :],
        "v_pol2": v_pol2[..., :-1, :],
        "freq": freqs,
    }


def main(simulation_path, destination_path):
    """
    This script fits spherical harmonics coefficients to a FEKO simulation for the OVRO-LWA telescope.
    The simulation file is provided as an HDF5 file.
    """

    # Order of Legendre polynomial to be fitted. Adjust to suit your needs
    nmax = 21

    """
    Simulated results for each frequency stored in separate file. Station coordinates
    stored in separate HDF5 file.
    """

    # Read simulated data from file
    simdata = read_lwa_simulation_file(simulation_path)

    fit_and_write_lobes_coeffs(
        simdata["theta"],
        simdata["phi"],
        simdata["freq"],
        None,
        simdata["v_pol1"],
        simdata["v_pol2"],
        nmax,
        destination_path,
        apply_weighting=False,
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Script to generate coeffiecients file for the OVRO-LWA telescope, based on a simulation file"
    )
    parser.add_argument(
        "--simulation_path", help="Path to LWA simulation file."
    )
    parser.add_argument(
        "--destination_path", help="Path to the output coefficients file."
    )
    args = parser.parse_args()
    main(args.simulation_path, args.destination_path)



#!/usr/bin/env python3
# Copyright (C) 2021 ASTRON (Netherlands Institute for Radio Astronomy)
# SPDX-License-Identifier: GPL-3.0-or-later

import everybeam as eb
import numpy as np

"""
Demo script illustrating how the EveryBeam Python
bindings can be used to compute beam responses.

Script requires an OSKAR Measurement Set in order to run.
A minimal OSKAR MSet can be downloaded here:
https://support.astron.nl/software/ci_data/EveryBeam/OSKAR-single-timeslot.tar.bz2
"""

# Set path to (OSKAR) MS
ms_path = os.path.join(os.environ["DATA_DIR"], "OSKAR_MOCK.ms")

# Use differential beam?
use_differential_beam = False

# Set element response model
element_response_model = "skala40_wave"

# Load the telescope
telescope = eb.load_telescope(
    ms_path,
    use_differential_beam=use_differential_beam,
    element_response_model=element_response_model,
)

# Is this an OSKAR telescope?
assert type(telescope) == eb.OSKAR

# Set properties for the beam evaluation
time = 4.45353e09
station_id = 0
freq = 5.0e07

# Point of interest (given in ITRF coordinates)
dir_itrf = np.array([-0.203137, 0.841818, -0.500078])

# Array factor for station 0, since we are evaluating at the phase centre (delay direction and
# direction of interest coincide), a unity matrix should be returned
array_factor_phase_centre = telescope.array_factor(
    time, station_id, freq, dir_itrf, dir_itrf
)
np.testing.assert_allclose(
    array_factor_phase_centre, np.eye(2, dtype=np.complex64), rtol=1e-8
)

# Just a rather arbitrarily chosen delay direction
delay_dir_itrf = dir_itrf + np.array([-0.1, 0.5, 0.025])
array_factor = telescope.array_factor(
    time, station_id, freq, dir_itrf, delay_dir_itrf
)

# Element beam for station 0
element_response = telescope.element_response(time, station_id, freq, dir_itrf)

# Full beam for station 0
response = telescope.station_response(
    time, station_id, freq, dir_itrf, delay_dir_itrf
)

# Full beam should match product of array_factor and element_response
np.testing.assert_allclose(
    np.matmul(array_factor, element_response), response, rtol=1e-6
)
